"""Photo service."""

import base64
import logging
import os
from abc import abstractmethod
from datetime import datetime, timezone
from enum import Enum, IntEnum, unique
from typing import Any, Generator, Iterable, Iterator, Optional, cast
from urllib.parse import urlencode

from requests import Response

from pyicloud.const import CONTENT_TYPE, CONTENT_TYPE_TEXT
from pyicloud.exceptions import (
    PyiCloudAPIResponseException,
    PyiCloudException,
    PyiCloudServiceNotActivatedException,
)
from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession

_LOGGER: logging.Logger = logging.getLogger(__name__)


class PhotosServiceException(PyiCloudException):
    """Photo service exception."""

    def __init__(
        self,
        *args,
        photo: "PhotoAsset|None" = None,
        album: "BasePhotoAlbum|None" = None,
    ) -> None:
        super().__init__(*args)
        self.photo: "PhotoAsset|None" = photo
        self.album: "BasePhotoAlbum|None" = album


@unique
class AlbumTypeEnum(IntEnum):
    """Album types"""

    ALBUM = 0
    FOLDER = 3
    SMART_ALBUM = 6


class SmartAlbumEnum(str, Enum):
    """Smart albums names."""

    ALL_PHOTOS = "Library"
    BURSTS = "Bursts"
    FAVORITES = "Favorites"
    HIDDEN = "Hidden"
    LIVE = "Live"
    PANORAMAS = "Panoramas"
    RECENTLY_DELETED = "Recently Deleted"
    SCREENSHOTS = "Screenshots"
    SLO_MO = "Slo-mo"
    TIME_LAPSE = "Time-lapse"
    VIDEOS = "Videos"


class DirectionEnum(str, Enum):
    """Direction names."""

    ASCENDING = "ASCENDING"
    DESCENDING = "DESCENDING"


class ListTypeEnum(str, Enum):
    """List type names."""

    DEFAULT = "CPLAssetAndMasterByAssetDateWithoutHiddenOrDeleted"
    DELETED = "CPLAssetAndMasterDeletedByExpungedDate"
    HIDDEN = "CPLAssetAndMasterHiddenByAssetDate"
    SMART_ALBUM = "CPLAssetAndMasterInSmartAlbumByAssetDate"
    STACK = "CPLBurstStackAssetAndMasterByAssetDate"
    CONTAINER = "CPLContainerRelationLiveByAssetDate"
    SHARED_STREAM = "sharedstream"


class ObjectTypeEnum(str, Enum):
    """Object type names."""

    ALL = "CPLAssetByAssetDateWithoutHiddenOrDeleted"
    BURST = "CPLAssetBurstStackAssetByAssetDate"
    DELETED = "CPLAssetDeletedByExpungedDate"
    FAVORITE = "CPLAssetInSmartAlbumByAssetDate:Favorite"
    HIDDEN = "CPLAssetHiddenByAssetDate"
    LIVE = "CPLAssetInSmartAlbumByAssetDate:Live"
    PANORAMA = "CPLAssetInSmartAlbumByAssetDate:Panorama"
    SCREENSHOT = "CPLAssetInSmartAlbumByAssetDate:Screenshot"
    SLOMO = "CPLAssetInSmartAlbumByAssetDate:Slomo"
    TIMELAPSE = "CPLAssetInSmartAlbumByAssetDate:Timelapse"
    VIDEO = "CPLAssetInSmartAlbumByAssetDate:Video"
    CONTAINER = "CPLContainerRelationNotDeletedByAssetDate"


# The primary zone for the user's photo library
PRIMARY_ZONE: dict[str, str] = {
    "zoneName": "PrimarySync",
    "zoneType": "REGULAR_CUSTOM_ZONE",
}


class AlbumContainer(Iterable):
    """
    Container for photo albums.
    This provides a way to access all the albums in the library.
    """

    def __init__(self, albums: list["BasePhotoAlbum"] | None = None) -> None:
        if albums is not None:
            self._albums: dict[str, "BasePhotoAlbum"] = {
                album.id: album for album in albums
            }
        else:
            self._albums = {}

        self._index: list[str] = list(self._albums.keys())

    def __len__(self) -> int:
        return len(self._albums)

    def __getitem__(self, key: str | int) -> "BasePhotoAlbum":
        """Returns the album for the given id."""
        if isinstance(key, int):
            return self._albums[self._index[key]]
        if key in self._albums:
            return self._albums[key]
        album: BasePhotoAlbum | None = self.find(key)
        if album is not None:
            return album
        raise KeyError(f"Photo album does not exist: {key}")

    def __iter__(self) -> Iterator["BasePhotoAlbum"]:
        return iter(self._albums.values())

    def __contains__(self, name: str) -> bool:
        """Checks if an album exists in the container."""
        return self.find(name) is not None

    def find(self, name: str) -> Optional["BasePhotoAlbum"]:
        """Finds an album by name, returns None if not found."""
        for album in self._albums.values():
            if name == album.fullname:
                return album
        return None

    def get(
        self, key: str, default: "BasePhotoAlbum | None" = None
    ) -> "BasePhotoAlbum | None":
        """Returns the album for the given key, or default if not found."""
        return self._albums.get(key, default)

    def append(self, album: "BasePhotoAlbum") -> None:
        """Appends an album to the container."""
        self._albums[album.id] = album
        self._index: list[str] = list(self._albums.keys())

    def index(self, idx: int) -> "BasePhotoAlbum":
        """Returns the album at the given index."""
        if idx < 0 or idx >= len(self._index):
            raise IndexError("Photo album index out of range")
        return self._albums[self._index[idx]]


class BasePhotoLibrary:
    """Represents a library in the user's photos.

    This provides access to all the albums as well as the photos.
    """

    def __init__(
        self,
        service: "PhotosService",
        asset_type: type["PhotoAsset"],
        upload_url: Optional[str] = None,
    ) -> None:
        self.service: PhotosService = service
        self.asset_type: type[PhotoAsset] = asset_type
        self._albums: Optional[AlbumContainer] = None
        self._upload_url: Optional[str] = upload_url

    @abstractmethod
    def _get_albums(self) -> AlbumContainer:
        """Returns the photo albums."""
        raise NotImplementedError

    @property
    def albums(self) -> AlbumContainer:
        """Returns the photo albums."""
        if self._albums is None:
            self._albums = self._get_albums()
        return self._albums

    def parse_asset_response(
        self, response: dict[str, list[dict[str, Any]]]
    ) -> tuple[dict[str, dict[str, Any]], list[dict[str, Any]]]:
        """Parses the asset response."""
        asset_records: dict[str, dict[str, Any]] = {}
        master_records: list[dict[str, Any]] = []
        for rec in response["records"]:
            if rec["recordType"] == "CPLAsset":
                master_id: str = rec["fields"]["masterRef"]["value"]["recordName"]
                asset_records[master_id] = rec
            elif rec["recordType"] == "CPLMaster":
                master_records.append(rec)
        return (asset_records, master_records)


class PhotoLibrary(BasePhotoLibrary):
    """Represents the user's primary photo libraries."""

    SMART_ALBUMS: dict[SmartAlbumEnum, dict[str, Any]] = {
        SmartAlbumEnum.ALL_PHOTOS: {
            "obj_type": ObjectTypeEnum.ALL,
            "list_type": ListTypeEnum.DEFAULT,
            "direction": DirectionEnum.DESCENDING,
            "query_filter": None,
        },
        SmartAlbumEnum.TIME_LAPSE: {
            "obj_type": ObjectTypeEnum.TIMELAPSE,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "TIMELAPSE"},
                }
            ],
        },
        SmartAlbumEnum.VIDEOS: {
            "obj_type": ObjectTypeEnum.VIDEO,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "VIDEO"},
                }
            ],
        },
        SmartAlbumEnum.SLO_MO: {
            "obj_type": ObjectTypeEnum.SLOMO,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "SLOMO"},
                }
            ],
        },
        SmartAlbumEnum.BURSTS: {
            "obj_type": ObjectTypeEnum.BURST,
            "list_type": ListTypeEnum.STACK,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": None,
        },
        SmartAlbumEnum.FAVORITES: {
            "obj_type": ObjectTypeEnum.FAVORITE,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "FAVORITE"},
                }
            ],
        },
        SmartAlbumEnum.PANORAMAS: {
            "obj_type": ObjectTypeEnum.PANORAMA,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "PANORAMA"},
                }
            ],
        },
        SmartAlbumEnum.SCREENSHOTS: {
            "obj_type": ObjectTypeEnum.SCREENSHOT,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "SCREENSHOT"},
                }
            ],
        },
        SmartAlbumEnum.LIVE: {
            "obj_type": ObjectTypeEnum.LIVE,
            "list_type": ListTypeEnum.SMART_ALBUM,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": [
                {
                    "fieldName": "smartAlbum",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "LIVE"},
                }
            ],
        },
        SmartAlbumEnum.RECENTLY_DELETED: {
            "obj_type": ObjectTypeEnum.DELETED,
            "list_type": ListTypeEnum.DELETED,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": None,
        },
        SmartAlbumEnum.HIDDEN: {
            "obj_type": ObjectTypeEnum.HIDDEN,
            "list_type": ListTypeEnum.HIDDEN,
            "direction": DirectionEnum.ASCENDING,
            "query_filter": None,
        },
    }

    def __init__(
        self,
        service: "PhotosService",
        zone_id: dict[str, str],
        upload_url: Optional[str] = None,
    ) -> None:
        super().__init__(service, asset_type=PhotoAsset, upload_url=upload_url)
        self.zone_id: dict[str, str] = zone_id

        self.url: str = (
            f"{self.service.service_endpoint}"
            f"/records/query?{urlencode(self.service.params)}"
        )
        request: Response = self.service.session.post(
            url=self.url,
            json={
                "query": {
                    "recordType": "CheckIndexingState",
                },
                "zoneID": self.zone_id,
            },
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        response: dict[str, Any] = request.json()
        indexing_state: str = response["records"][0]["fields"]["state"]["value"]
        if indexing_state != "FINISHED":
            _LOGGER.debug("iCloud Photo Library not finished indexing")
            raise PyiCloudServiceNotActivatedException(
                "iCloud Photo Library not finished indexing. "
                "Please try again in a few minutes."
            )

    def _fetch_records(self, parent_id: Optional[str] = None) -> list[dict[str, Any]]:
        """Fetches records."""
        query: dict[str, Any] = {
            "query": {
                "recordType": "CPLAlbumByPositionLive",
            },
            "zoneID": self.zone_id,
        }

        if parent_id:
            query["query"]["filterBy"] = [
                {
                    "fieldName": "parentId",
                    "comparator": "EQUALS",
                    "fieldValue": {
                        "value": parent_id,
                        "type": "STRING",
                    },
                }
            ]

        request: Response = self.service.session.post(
            url=self.url,
            json=query,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        response: dict[str, list[dict[str, Any]]] = request.json()
        records: list[dict[str, Any]] = response["records"]

        while "continuationMarker" in response:
            query["continuationMarker"] = response["continuationMarker"]

            request: Response = self.service.session.post(
                url=self.url,
                json=query,
                headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            )
            response = request.json()
            records.extend(response["records"])

        for record in records.copy():
            if (
                record["fields"].get("albumType")
                and record["fields"]["albumType"]["value"] == AlbumTypeEnum.FOLDER.value
            ):
                records.extend(self._fetch_records(parent_id=record["recordName"]))

        return records

    def _convert_record_to_album(
        self, record: dict[str, Any]
    ) -> Optional["PhotoAlbum"]:
        """Converts a record to a photo album."""
        if (
            # Skipping albums having null name, that can happen sometime
            "albumNameEnc" not in record["fields"]
            or (
                record["fields"].get("isDeleted")
                and record["fields"]["isDeleted"]["value"]
            )
        ):
            return None

        record_id: str = record["recordName"]
        album_name: str = base64.b64decode(
            record["fields"]["albumNameEnc"]["value"]
        ).decode("utf-8")

        query_filter: list[dict[str, Any]] = [
            {
                "fieldName": "parentId",
                "comparator": "EQUALS",
                "fieldValue": {"type": "STRING", "value": record_id},
            }
        ]

        parent_id: Optional[str] = record["fields"].get("parentId", {}).get("value")

        album_type: type[PhotoAlbum] = PhotoAlbum

        if (
            record["fields"].get("albumType")
            and record["fields"]["albumType"]["value"] == AlbumTypeEnum.FOLDER.value
        ):
            album_type = PhotoAlbumFolder

        direction: DirectionEnum = DirectionEnum.ASCENDING
        if record["fields"].get("sortAscending", {}).get("value", 1) != 1:
            direction = DirectionEnum.DESCENDING

        record_modification_date = (
            record["fields"].get("recordModificationDate", {}).get("value", None)
        )

        return album_type(
            library=self,
            name=album_name,
            record_id=record_id,
            list_type=ListTypeEnum.CONTAINER,
            obj_type=ObjectTypeEnum.CONTAINER,
            direction=direction,
            url=self.url,
            query_filter=query_filter,
            zone_id=record.get("zoneID", self.zone_id),
            parent_id=parent_id,
            record_change_tag=record["recordChangeTag"],
            record_modification_date=record_modification_date,
        )

    def _get_albums(self) -> AlbumContainer:
        """Returns photo albums."""
        albums = AlbumContainer(
            [
                SmartPhotoAlbum(
                    library=self,
                    name=name,
                    zone_id=self.zone_id,
                    url=self.url,
                    **props,
                )
                for (name, props) in self.SMART_ALBUMS.items()
            ]
        )

        for record in self._fetch_records():
            album: PhotoAlbum | None = self._convert_record_to_album(record)
            if album is not None:
                albums.append(album)

        return albums

    def create_album(
        self, name: str, album_type: AlbumTypeEnum = AlbumTypeEnum.ALBUM
    ) -> Optional["PhotoAlbum"]:
        """Creates a new album, returns the request response."""
        data: dict[str, Any] = {
            "operations": [
                {
                    "operationType": "create",
                    "record": {
                        "recordType": "CPLAlbum",
                        "fields": {
                            "albumNameEnc": {
                                "value": base64.b64encode(name.encode("utf-8")).decode(
                                    "utf-8"
                                ),
                            },
                            "albumType": {
                                "value": album_type.value,
                            },
                            "isDeleted": {
                                "value": 0,
                            },
                            "isExpunged": {
                                "value": 0,
                            },
                            "sortType": {
                                "value": 1,
                            },
                            "sortAscending": {
                                "value": 1,
                            },
                        },
                    },
                }
            ],
            "zoneID": self.zone_id,
            "atomic": True,
        }

        endpoint: str = self.service.service_endpoint
        params: str = urlencode(self.service.params)
        url: str = f"{endpoint}/records/modify?{params}"

        try:
            resp: Response = self.service.session.post(
                url,
                json=data,
                headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            )

            payload: dict[str, Any] = resp.json()
            records: list[dict[str, Any]] = payload.get("records", [])
            if not records:
                return None
        except PyiCloudAPIResponseException as ex:
            _LOGGER.error("Failed to create album: %s", ex)
            raise PhotosServiceException("Failed to create album") from ex

        return self._convert_record_to_album(records[0])

    def upload_file(self, path: str) -> Optional["PhotoAsset"]:
        """Upload a photo from path, returns a recordName"""

        filename: str = os.path.basename(path)

        params: dict[str, Any] = self.service.params.copy()
        params["filename"] = filename

        url: str = f"{self._upload_url}/upload?{urlencode(params)}"

        with open(path, "rb") as file_obj:
            response: Response = self.service.session.post(
                url=url,
                data=file_obj,
            )

        json_response: dict[str, Any] = response.json()
        if "errors" in json_response:
            raise PyiCloudAPIResponseException("", json_response["errors"])

        records: dict[Any, dict[str, Any]] = {
            rec["recordType"]: rec for rec in json_response["records"]
        }

        if "CPLMaster" not in records or "CPLAsset" not in records:
            return None

        return self.asset_type(self.service, records["CPLMaster"], records["CPLAsset"])

    @property
    def all(self) -> "PhotoAlbum":
        """Returns the All Photos album."""
        return cast(PhotoAlbum, self.albums[SmartAlbumEnum.ALL_PHOTOS])


class PhotoStreamLibrary(BasePhotoLibrary):
    """Represents a shared photo library."""

    def __init__(
        self,
        service: "PhotosService",
        shared_streams_url: str,
    ) -> None:
        super().__init__(service, asset_type=PhotoStreamAsset, upload_url=None)
        self.shared_streams_url: str = shared_streams_url

    def _get_albums(self) -> AlbumContainer:
        """Returns albums."""
        albums: AlbumContainer = AlbumContainer()
        url: str = f"{self.shared_streams_url}?{urlencode(self.service.params)}"
        request: Response = self.service.session.post(
            url, json={}, headers={CONTENT_TYPE: CONTENT_TYPE_TEXT}
        )
        response: dict[str, list] = request.json()
        for album in response["albums"]:
            shared_stream = SharedPhotoStreamAlbum(
                library=self,
                name=album["attributes"]["name"],
                album_location=album["albumlocation"],
                album_ctag=album["albumctag"],
                album_guid=album["albumguid"],
                owner_dsid=album["ownerdsid"],
                creation_date=album["attributes"]["creationDate"],
                sharing_type=album["sharingtype"],
                allow_contributions=album["attributes"]["allowcontributions"],
                is_public=album["attributes"]["ispublic"],
                is_web_upload_supported=album["iswebuploadsupported"],
                public_url=album.get("publicurl", None),
            )
            albums.append(shared_stream)
        return albums


class PhotosService(BaseService):
    """The 'Photos' iCloud service.

    This also acts as a way to access the user's primary library."""

    def __init__(
        self,
        service_root: str,
        session: PyiCloudSession,
        params: dict[str, Any],
        upload_url: str,
        shared_streams_url: str,
    ) -> None:
        BaseService.__init__(
            self,
            service_root=service_root,
            session=session,
            params=params,
        )
        self.service_endpoint: str = (
            f"{self.service_root}/database/1/com.apple.photos.cloud/production/private"
        )

        self._libraries: Optional[dict[str, BasePhotoLibrary]] = None

        self.params.update({"remapEnums": True, "getCurrentSyncToken": True})
        self._photo_assets: dict = {}

        self._root_library: PhotoLibrary = PhotoLibrary(
            self,
            PRIMARY_ZONE,
            upload_url=upload_url,
        )

        self._shared_library: PhotoStreamLibrary = PhotoStreamLibrary(
            self,
            shared_streams_url=(
                f"{shared_streams_url}/{self.params['dsid']}"
                "/sharedstreams/webgetalbumslist"
            ),
        )

    @property
    def libraries(self) -> dict[str, BasePhotoLibrary]:
        """Returns photo libraries."""
        if not self._libraries:
            url: str = f"{self.service_endpoint}/changes/database"

            request: Response = self.session.post(
                url, data="{}", headers={CONTENT_TYPE: CONTENT_TYPE_TEXT}
            )
            response: dict[str, Any] = request.json()
            zones: list[dict[str, Any]] = response["zones"]

            libraries: dict[str, BasePhotoLibrary] = {
                "root": self._root_library,
                "shared": self._shared_library,
            }
            for zone in zones:
                if not zone.get("deleted"):
                    zone_name: str = zone["zoneID"]["zoneName"]
                    libraries[zone_name] = PhotoLibrary(self, zone["zoneID"])

            self._libraries = libraries

        return self._libraries

    @property
    def all(self) -> "PhotoAlbum":
        """Returns the primary photo library."""
        return self._root_library.all

    @property
    def albums(self) -> AlbumContainer:
        """Returns the standard photo albums."""
        return self._root_library.albums

    @property
    def shared_streams(self) -> AlbumContainer:
        """Returns the shared photo albums."""
        return self._shared_library.albums

    def create_album(
        self, name: str, album_type: AlbumTypeEnum = AlbumTypeEnum.ALBUM
    ) -> Optional["PhotoAlbum"]:
        """Creates a new album in the primary photo library."""
        return self._root_library.create_album(name, album_type)


class BasePhotoAlbum:
    """An abstract photo album."""

    def __init__(
        self,
        library: BasePhotoLibrary,
        name: str,
        list_type: ListTypeEnum,
        page_size: int = 100,
        direction: DirectionEnum = DirectionEnum.ASCENDING,
    ) -> None:
        self._name: str = name
        self._library: BasePhotoLibrary = library
        self._page_size: int = page_size
        self._direction: DirectionEnum = direction
        self._list_type: ListTypeEnum = list_type
        self._len: Optional[int] = None

    @property
    @abstractmethod
    def fullname(self) -> str:
        """Gets the full name of the album including path"""
        raise NotImplementedError

    @property
    def page_size(self) -> int:
        """Gets the page size."""
        return self._page_size if self._page_size < 100 else 100

    @property
    def service(self) -> PhotosService:
        """Get the Photo service"""
        return self._library.service

    def _get_photos_at(
        self, index: int, direction: DirectionEnum, page_size: int
    ) -> Generator["PhotoAsset", None, None]:
        offset: int = max(0, index)

        response: Response = self.service.session.post(
            url=self._get_url(),
            json=self._get_payload(
                offset=offset,
                page_size=page_size
                * 2,  # Fetch double the page size to cater for master and asset records
                direction=direction,
            ),
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        json_response: dict[str, list[dict[str, Any]]] = response.json()
        asset_records: dict[str, Any]
        master_records: list[dict[str, Any]]
        asset_records, master_records = self._library.parse_asset_response(
            json_response
        )
        for master_record in master_records:
            record_name: str = master_record["recordName"]
            asset_record = asset_records.get(record_name)
            if not asset_record:
                _LOGGER.debug(
                    "No asset record found for master record: %s", record_name
                )
                continue
            yield self._library.asset_type(self.service, master_record, asset_record)

    def photo(self, index) -> "PhotoAsset":
        """Returns a photo at the given index."""
        return next(self._get_photos_at(index, self._direction, 1))

    @property
    def title(self) -> str:
        """Gets the album title."""
        return self.name

    @property
    def name(self) -> str:
        """Gets the album name."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Sets the album name."""
        if self._name != value:
            self.rename(value)

    def rename(self, value: str) -> None:
        """Renames the album."""
        raise NotImplementedError("Album name is read-only")

    def delete(self) -> bool:
        """Deletes the album."""
        raise NotImplementedError("Album delete is not implemented")

    @property
    def photos(self) -> Generator["PhotoAsset", None, None]:
        """Returns the album photos."""
        self._len = None
        if self._direction == DirectionEnum.DESCENDING:
            offset: int = len(self) - 1
        else:
            offset = 0

        photos_ids: set[str] = set()

        while True:
            num_results = 0
            for photo in self._get_photos_at(offset, self._direction, self.page_size):
                num_results += 1
                if photo.id in photos_ids:
                    _LOGGER.debug("Duplicate photo found: %s, skipping", photo.id)
                    continue
                photos_ids.add(photo.id)
                yield photo
            if num_results < self.page_size:
                _LOGGER.debug("Less than page size returned: %d", num_results)
            if (
                num_results < self.page_size // 2
            ):  # If less than half the page size is returned, we assume we're done
                break
            if self._direction == DirectionEnum.DESCENDING:
                offset = offset - num_results
            else:
                offset = offset + num_results

    @property
    @abstractmethod
    def id(self) -> str:
        """Gets the album id."""
        raise NotImplementedError

    @abstractmethod
    def _get_payload(
        self, offset: int, page_size: int, direction: DirectionEnum
    ) -> dict[str, Any]:
        """Returns the payload for the photo list request."""
        raise NotImplementedError

    @abstractmethod
    def _get_url(self) -> str:
        """Returns the URL for the photo list request."""
        raise NotImplementedError

    @abstractmethod
    def _get_len(self) -> int:
        """Returns the number of photos in the album."""
        raise NotImplementedError

    def __iter__(self) -> Generator["PhotoAsset", None, None]:
        return self.photos

    def __len__(self) -> int:
        if self._len is None:
            self._len = self._get_len()
        return self._len

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: '{self}'>"


class PhotoAlbum(BasePhotoAlbum):
    """A photo album."""

    def __init__(
        self,
        library: PhotoLibrary,
        name: str,
        record_id: str,
        obj_type: ObjectTypeEnum,
        list_type: ListTypeEnum,
        direction: DirectionEnum,
        url: str,
        query_filter: Optional[list[dict[str, Any]]] = None,
        zone_id: Optional[dict[str, str]] = None,
        page_size: int = 100,
        parent_id: Optional[str] = None,
        record_change_tag: Optional[str] = None,
        record_modification_date: Optional[str] = None,
    ) -> None:
        super().__init__(
            library=library,
            name=name,
            list_type=list_type,
            page_size=page_size,
            direction=direction,
        )

        self._record_id: str = record_id
        self._obj_type: ObjectTypeEnum = obj_type
        self._query_filter: Optional[list[dict[str, Any]]] = query_filter
        self._url: str = url
        self._parent_id: Optional[str] = parent_id
        self._record_change_tag: Optional[str] = record_change_tag
        self._record_modification_date: Optional[str] = record_modification_date

        if zone_id:
            self._zone_id: dict[str, str] = zone_id
        else:
            self._zone_id = PRIMARY_ZONE

    @property
    def id(self) -> str:
        """Gets the album id."""
        return self._record_id

    @property
    def fullname(self) -> str:
        if self._parent_id is not None:
            return f"{self._library.albums[self._parent_id].fullname}/{self.name}"

        return self.name

    def rename(self, value: str) -> None:
        """Renames the album."""
        if self._name == value:
            return

        data: dict[str, Any] = {
            "atomic": True,
            "zoneID": self._zone_id,
            "operations": [
                {
                    "operationType": "update",
                    "record": {
                        "recordName": self._record_id,
                        "recordType": "CPLAlbum",
                        "recordChangeTag": self._record_change_tag,
                        "fields": {
                            "albumNameEnc": {
                                "value": base64.b64encode(value.encode("utf-8")).decode(
                                    "utf-8"
                                ),
                            },
                        },
                    },
                }
            ],
        }
        url: str = (
            f"{self.service.service_endpoint}/records/modify"
            f"?{urlencode(self.service.params)}"
        )

        response: Response = self.service.session.post(
            url,
            json=data,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        payload: dict[str, Any] = response.json()
        if payload.get("records"):
            latest: dict[str, Any] = payload["records"][0]
            self._record_change_tag = latest.get(
                "recordChangeTag", self._record_change_tag
            )
            fields: dict[str, Any] = latest.get("fields", {})
            self._record_modification_date = fields.get(
                "recordModificationDate", {}
            ).get("value", self._record_modification_date)

        self._name = value

    def delete(self) -> bool:
        """Deletes the album."""
        data: dict[str, Any] = {
            "atomic": True,
            "zoneID": self._zone_id,
            "operations": [
                {
                    "operationType": "update",
                    "record": {
                        "recordName": self._record_id,
                        "recordChangeTag": self._record_change_tag,
                        "recordType": "CPLAlbum",
                        "fields": {
                            "isDeleted": {"value": 1},
                        },
                    },
                }
            ],
        }
        url: str = (
            f"{self.service.service_endpoint}/records/modify"
            f"?{urlencode(self.service.params)}"
        )

        try:
            response: Response = self.service.session.post(
                url,
                json=data,
                headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            )
            payload: dict[str, Any] = response.json()
            self._record_change_tag = payload["records"][0].get(
                "recordChangeTag", self._record_change_tag
            )
            self._record_modification_date = (
                payload["records"][0]
                .get("fields", {})
                .get("recordModificationDate", {})
                .get("value", self._record_modification_date)
            )
        except PyiCloudAPIResponseException as ex:
            _LOGGER.error("Failed to delete photo from album: %s", ex)
            raise PhotosServiceException(
                "Failed to delete photo from album", album=self
            ) from ex

        return True

    def add_photo(self, photo: "PhotoAsset") -> bool:
        """Adds an existing photo to the album."""

        data: dict[str, Any] = {
            "atomic": True,
            "zoneID": self._zone_id,
            "operations": [
                {
                    "operationType": "create",
                    "record": {
                        "fields": {
                            "itemId": {"value": photo.id},
                            "position": {"value": 1024},
                            "containerId": {"value": self._record_id},
                        },
                        "recordType": "CPLContainerRelation",
                        "recordName": f"{photo.id}-IN-{self._record_id}",
                    },
                }
            ],
        }
        url: str = (
            f"{self.service.service_endpoint}/records/modify"
            f"?{urlencode(self.service.params)}"
        )

        try:
            response: Response = self.service.session.post(
                url,
                json=data,
                headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            )

            payload: dict[str, Any] = response.json()
            self._record_change_tag = payload["records"][0].get(
                "recordChangeTag", self._record_change_tag
            )
            self._record_modification_date = (
                payload["records"][0]
                .get("fields", {})
                .get("recordModificationDate", {})
                .get("value", self._record_modification_date)
            )
        except PyiCloudAPIResponseException as ex:
            _LOGGER.error("Failed to add photo to album: %s", ex)
            return False

        return True

    def upload(self, path) -> Optional["PhotoAsset"]:
        """Uploads a photo to the album."""
        if not isinstance(self._library, PhotoLibrary):
            return None
        photo_asset: PhotoAsset | None = self._library.upload_file(path)

        if photo_asset is None:
            return None

        if not self.add_photo(photo_asset):
            _LOGGER.error("Failed to add photo to album")
            raise PhotosServiceException(
                "Failed to add photo to album",
                album=self,
                photo=photo_asset,
            )

        return photo_asset

    @property
    def _get_container_id(self) -> str:
        """Returns the container ID."""
        return f"{self._obj_type.value}:{self._record_id}"

    def _get_len(self) -> int:
        url: str = (
            f"{self.service.service_endpoint}/internal/records/query/batch"
            f"?{urlencode(self.service.params)}"
        )
        request: Response = self.service.session.post(
            url,
            json={
                "batch": [
                    {
                        "resultsLimit": 1,
                        "query": {
                            "recordType": "HyperionIndexCountLookup",
                            "filterBy": {
                                "fieldName": "indexCountID",
                                "comparator": "IN",
                                "fieldValue": {
                                    "type": "STRING_LIST",
                                    "value": [self._get_container_id],
                                },
                            },
                        },
                        "zoneWide": True,
                        "zoneID": self._zone_id,
                    }
                ]
            },
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        response: dict[str, Any] = request.json()

        return response["batch"][0]["records"][0]["fields"]["itemCount"]["value"]

    def _get_payload(
        self, offset: int, page_size: int, direction: DirectionEnum
    ) -> dict[str, Any]:
        return self._list_query_gen(
            offset,
            self._list_type,
            direction,
            page_size,
            self._query_filter,
        )

    def _get_url(self) -> str:
        return self._url

    def _list_query_gen(
        self,
        offset: int,
        list_type: ListTypeEnum,
        direction: DirectionEnum,
        num_results: int,
        query_filter=None,
    ) -> dict[str, Any]:
        query: dict[str, Any] = {
            "query": {
                "recordType": list_type.value,
                "filterBy": [
                    {
                        "fieldName": "direction",
                        "comparator": "EQUALS",
                        "fieldValue": {"type": "STRING", "value": direction.value},
                    },
                    {
                        "fieldName": "startRank",
                        "comparator": "EQUALS",
                        "fieldValue": {"type": "INT64", "value": offset},
                    },
                ],
            },
            "resultsLimit": num_results,
            "desiredKeys": [
                "resJPEGFullWidth",
                "resJPEGFullHeight",
                "resJPEGFullFileType",
                "resJPEGFullFingerprint",
                "resJPEGFullRes",
                "resJPEGLargeWidth",
                "resJPEGLargeHeight",
                "resJPEGLargeFileType",
                "resJPEGLargeFingerprint",
                "resJPEGLargeRes",
                "resJPEGMedWidth",
                "resJPEGMedHeight",
                "resJPEGMedFileType",
                "resJPEGMedFingerprint",
                "resJPEGMedRes",
                "resJPEGThumbWidth",
                "resJPEGThumbHeight",
                "resJPEGThumbFileType",
                "resJPEGThumbFingerprint",
                "resJPEGThumbRes",
                "resVidFullWidth",
                "resVidFullHeight",
                "resVidFullFileType",
                "resVidFullFingerprint",
                "resVidFullRes",
                "resVidMedWidth",
                "resVidMedHeight",
                "resVidMedFileType",
                "resVidMedFingerprint",
                "resVidMedRes",
                "resVidSmallWidth",
                "resVidSmallHeight",
                "resVidSmallFileType",
                "resVidSmallFingerprint",
                "resVidSmallRes",
                "resSidecarWidth",
                "resSidecarHeight",
                "resSidecarFileType",
                "resSidecarFingerprint",
                "resSidecarRes",
                "itemType",
                "dataClassType",
                "filenameEnc",
                "originalOrientation",
                "resOriginalWidth",
                "resOriginalHeight",
                "resOriginalFileType",
                "resOriginalFingerprint",
                "resOriginalRes",
                "resOriginalAltWidth",
                "resOriginalAltHeight",
                "resOriginalAltFileType",
                "resOriginalAltFingerprint",
                "resOriginalAltRes",
                "resOriginalVidComplWidth",
                "resOriginalVidComplHeight",
                "resOriginalVidComplFileType",
                "resOriginalVidComplFingerprint",
                "resOriginalVidComplRes",
                "isDeleted",
                "isExpunged",
                "dateExpunged",
                "remappedRef",
                "recordName",
                "recordType",
                "recordChangeTag",
                "masterRef",
                "adjustmentRenderType",
                "assetDate",
                "addedDate",
                "isFavorite",
                "isHidden",
                "orientation",
                "duration",
                "assetSubtype",
                "assetSubtypeV2",
                "assetHDRType",
                "burstFlags",
                "burstFlagsExt",
                "burstId",
                "captionEnc",
                "locationEnc",
                "locationV2Enc",
                "locationLatitude",
                "locationLongitude",
                "adjustmentType",
                "timeZoneOffset",
                "vidComplDurValue",
                "vidComplDurScale",
                "vidComplDispValue",
                "vidComplDispScale",
                "vidComplVisibilityState",
                "customRenderedValue",
                "containerId",
                "itemId",
                "position",
                "isKeyAsset",
            ],
            "zoneID": self._zone_id,
        }

        if query_filter:
            query["query"]["filterBy"].extend(query_filter)

        return query


class PhotoAlbumFolder(PhotoAlbum):
    """A Photo Album Folder."""

    def upload(self, path) -> Optional["PhotoAsset"]:
        """Uploads a photo to the album."""
        # Folders do not support uploads
        return None


class SmartPhotoAlbum(PhotoAlbum):
    """A Smart Photo Album."""

    def __init__(
        self,
        library: PhotoLibrary,
        name: SmartAlbumEnum,
        obj_type: ObjectTypeEnum,
        list_type: ListTypeEnum,
        direction: DirectionEnum,
        url: str,
        query_filter: Optional[list[dict[str, Any]]] = None,
        zone_id: Optional[dict[str, str]] = None,
        page_size: int = 100,
        parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            library=library,
            name=name.value,
            record_id=name.value,
            obj_type=obj_type,
            list_type=list_type,
            direction=direction,
            url=url,
            query_filter=query_filter,
            zone_id=zone_id,
            page_size=page_size,
            parent_id=parent_id,
        )

    @property
    def id(self) -> str:
        """Gets the album id."""
        return self.name

    def upload(self, path) -> Optional["PhotoAsset"]:
        """Uploads a photo to the album."""
        # Smart albums do not support uploads
        return None

    @property
    def fullname(self) -> str:
        """Gets the full name of the album including path"""
        return self.name

    @property
    def _get_container_id(self) -> str:
        """Gets the container ID."""
        return f"{self._obj_type.value}"


class SharedPhotoStreamAlbum(BasePhotoAlbum):
    """A Shared Stream Photo Album."""

    def __init__(
        self,
        library: BasePhotoLibrary,
        name: str,
        album_location: str,
        album_ctag: str,
        album_guid: str,
        owner_dsid: str,
        creation_date: str,
        sharing_type: str = "owned",
        allow_contributions: bool = False,
        is_public: bool = False,
        is_web_upload_supported: bool = False,
        public_url: Optional[str] = None,
        page_size: int = 100,
    ) -> None:
        super().__init__(
            library=library,
            name=name,
            list_type=ListTypeEnum.SHARED_STREAM,
            page_size=page_size,
        )

        self._album_location: str = album_location
        self._album_ctag: str = album_ctag
        self._album_guid: str = album_guid
        self._owner_dsid: str = owner_dsid
        try:
            self.creation_date: datetime = datetime.fromtimestamp(
                int(creation_date) / 1000.0, timezone.utc
            )
        except ValueError:
            self.creation_date = datetime.fromtimestamp(0, timezone.utc)

        # Read only properties
        self._sharing_type: str = sharing_type
        self._allow_contributions: bool = allow_contributions
        self._is_public: bool = is_public
        self._is_web_upload_supported: bool = is_web_upload_supported
        self._public_url: Optional[str] = public_url

    @property
    def id(self) -> str:
        """Gets the album id."""
        return self._album_guid

    @property
    def fullname(self) -> str:
        return self.name

    @property
    def sharing_type(self) -> str:
        """Gets the sharing type."""
        return self._sharing_type

    @property
    def allow_contributions(self) -> bool:
        """Gets if contributions are allowed."""
        return self._allow_contributions

    @property
    def is_public(self) -> bool:
        """Gets if the album is public."""
        return self._is_public

    @property
    def is_web_upload_supported(self) -> bool:
        """Gets if web uploads are supported."""
        return self._is_web_upload_supported

    @property
    def public_url(self) -> Optional[str]:
        """Gets the public URL."""
        return self._public_url

    def _get_payload(
        self, offset: int, page_size: int, direction: DirectionEnum
    ) -> dict[str, Any]:
        return {
            "albumguid": self._album_guid,
            "albumctag": self._album_ctag,
            "limit": str(min(offset + page_size, len(self))),
            "offset": str(offset),
        }

    def _get_url(self) -> str:
        return f"{self._album_location}webgetassets?{urlencode(self.service.params)}"

    def _get_len(self) -> int:
        url: str = (
            f"{self._album_location}webgetassetcount?{urlencode(self.service.params)}"
        )
        request: Response = self.service.session.post(
            url,
            json={
                "albumguid": self._album_guid,
            },
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        response: dict[str, Any] = request.json()

        return response["albumassetcount"]

    def delete(self) -> bool:
        """Deletes the album."""
        # Shared albums cannot be deleted
        return False

    def rename(self, value: str) -> None:
        """Renames the album."""
        # Shared albums cannot be renamed
        return None


class PhotoAsset:
    """A photo."""

    def __init__(
        self,
        service: PhotosService,
        master_record: dict[str, Any],
        asset_record: dict[str, Any],
    ) -> None:
        self._service: PhotosService = service
        self._master_record: dict[str, Any] = master_record
        self._asset_record: dict[str, Any] = asset_record

        self._versions: Optional[dict[str, dict[str, Any]]] = None

    ITEM_TYPES: dict[str, str] = {
        "public.heic": "image",
        "public.jpeg": "image",
        "public.png": "image",
        "com.apple.quicktime-movie": "movie",
    }

    FILE_TYPE_EXTENSIONS: dict[str, str] = {
        "public.heic": ".HEIC",
        "public.jpeg": ".JPG",
        "public.png": ".PNG",
        "com.apple.quicktime-movie": ".MOV",
    }

    PHOTO_VERSION_LOOKUP: dict[str, str] = {
        "original": "resOriginal",
        "medium": "resJPEGMed",
        "thumb": "resJPEGThumb",
        "original_video": "resOriginalVidCompl",
        "medium_video": "resVidMed",
        "thumb_video": "resVidSmall",
    }

    VIDEO_VERSION_LOOKUP: dict[str, str] = {
        "original": "resOriginal",
        "medium": "resVidMed",
        "thumb": "resVidSmall",
    }

    @property
    def id(self) -> str:
        """Gets the photo id."""
        return self._master_record["recordName"]

    @property
    def filename(self) -> str:
        """Gets the photo file name."""
        return base64.b64decode(
            self._master_record["fields"]["filenameEnc"]["value"]
        ).decode("utf-8")

    @property
    def size(self):
        """Gets the photo size."""
        return self._master_record["fields"]["resOriginalRes"]["value"]["size"]

    @property
    def created(self) -> datetime:
        """Gets the photo created date."""
        return self.asset_date

    @property
    def asset_date(self) -> datetime:
        """Gets the photo asset date."""
        try:
            return datetime.fromtimestamp(
                self._asset_record["fields"]["assetDate"]["value"] / 1000.0,
                timezone.utc,
            )
        except KeyError:
            return datetime.fromtimestamp(0, timezone.utc)

    @property
    def added_date(self) -> datetime:
        """Gets the photo added date."""
        return datetime.fromtimestamp(
            self._asset_record["fields"]["addedDate"]["value"] / 1000.0, timezone.utc
        )

    @property
    def dimensions(self):
        """Gets the photo dimensions."""
        return (
            self._master_record["fields"]["resOriginalWidth"]["value"],
            self._master_record["fields"]["resOriginalHeight"]["value"],
        )

    @property
    def item_type(self) -> str:
        """Gets the photo item type."""
        item_type: str = ""
        try:
            item_type = self._master_record["fields"]["itemType"]["value"]
        except KeyError:
            try:
                item_type = self._master_record["fields"]["resOriginalFileType"][
                    "value"
                ]
            except KeyError:
                # Both fields missing; fall back to filename extension or default to "movie".
                pass
        if item_type in self.ITEM_TYPES:
            return self.ITEM_TYPES[item_type]
        if self.filename.lower().endswith((".heic", ".png", ".jpg", ".jpeg")):
            return "image"
        return "movie"

    @property
    def is_live_photo(self) -> bool:
        """Check if the photo is a live photo."""
        return (
            self.item_type == "image"
            and "resOriginalVidComplFileType" in self._master_record["fields"]
        )

    @property
    def versions(self) -> dict[str, dict[str, Any]]:
        """Gets the photo versions."""
        if not self._versions:
            self._versions = {}
            if self.item_type == "movie":
                typed_version_lookup: dict[str, str] = self.VIDEO_VERSION_LOOKUP
            else:
                typed_version_lookup = self.PHOTO_VERSION_LOOKUP

            for key, prefix in typed_version_lookup.items():
                if f"{prefix}Res" in self._master_record["fields"]:
                    self._versions[key] = self._get_photo_version(prefix)

        return self._versions

    def _get_photo_version(self, prefix: str) -> dict[str, Any]:
        version: dict[str, Any] = {}
        fields: dict[str, dict[str, Any]] = self._master_record["fields"]
        width_entry: Optional[dict[str, Any]] = fields.get(f"{prefix}Width")
        if width_entry:
            version["width"] = width_entry["value"]
        else:
            version["width"] = None

        height_entry: Optional[dict[str, Any]] = fields.get(f"{prefix}Height")
        if height_entry:
            version["height"] = height_entry["value"]
        else:
            version["height"] = None

        size_entry: Optional[dict[str, Any]] = fields.get(f"{prefix}Res")
        if size_entry:
            version["size"] = size_entry["value"]["size"]
            version["url"] = size_entry["value"]["downloadURL"]
        else:
            version["size"] = None
            version["url"] = None

        type_entry: Optional[dict[str, Any]] = fields.get(f"{prefix}FileType")
        if type_entry:
            version["type"] = type_entry["value"]
        else:
            version["type"] = None

        # Default to the master filename.
        version["filename"] = self.filename
        # For live photos, the video version has a different filename.
        if self.is_live_photo:
            version_type: Optional[str] = version.get("type")
            # Check if the current version is the video component of the live photo.
            if version_type and self.ITEM_TYPES.get(version_type, None) == "movie":
                # Create the video filename from the image filename.
                # e.g. IMG_1234.HEIC -> IMG_1234.MOV
                filename_base, _ = os.path.splitext(self.filename)
                extension: str = self.FILE_TYPE_EXTENSIONS.get(version_type, ".MOV")
                live_photo_video_filename: str = f"{filename_base}{extension}"
                version["filename"] = live_photo_video_filename

        return version

    def download(self, version="original", **kwargs) -> Optional[bytes]:
        """Returns the photo file."""
        if version not in self.versions:
            return None

        response: Response = self._service.session.get(
            self.versions[version]["url"],
            stream=True,
            **kwargs,
        )
        return response.raw.read()

    def delete(self) -> bool:
        """Deletes the photo."""
        endpoint: str = self._service.service_endpoint
        params: str = urlencode(self._service.params)
        url: str = f"{endpoint}/records/modify?{params}"

        resp: Response = self._service.session.post(
            url,
            json={
                "operations": [
                    {
                        "operationType": "update",
                        "record": {
                            "recordName": self._asset_record["recordName"],
                            "recordType": self._asset_record["recordType"],
                            "recordChangeTag": self._asset_record.get(
                                "recordChangeTag",
                                self._master_record.get("recordChangeTag"),
                            ),
                            "fields": {"isDeleted": {"value": 1}},
                        },
                    }
                ],
                "zoneID": self._asset_record["zoneID"],
                "atomic": True,
            },
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
        )
        return resp.status_code == 200

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: id={self.id}>"


class PhotoStreamAsset(PhotoAsset):
    """A Shared Stream Photo Asset"""

    @property
    def like_count(self) -> int:
        """Gets the photo like count."""
        return (
            self._asset_record.get("pluginFields", {})
            .get("likeCount", {})
            .get("value", 0)
        )

    @property
    def liked(self) -> bool:
        """Gets if the photo is liked."""
        return bool(
            self._asset_record.get("pluginFields", {})
            .get("likedByCaller", {})
            .get("value", False)
        )



"""Test calendar service"""
# pylint: disable=protected-access

from datetime import datetime
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from requests import Response

from pyicloud.services.calendar import (
    AlarmDefaults,
    AlarmMeasurement,
    AppleAlarm,
    AppleDateFormat,
    CalendarDefaults,
    CalendarObject,
    CalendarService,
    DateFormats,
    EventObject,
)
from pyicloud.session import PyiCloudSession


def test_event_object_initialization() -> None:
    """Test EventObject initialization and default values."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="calendar123")
        assert event.pguid == "calendar123"
        assert event.title == "New Event"
        assert event.duration == 60
        assert event.tz == "UTC"  # Now tests dynamic timezone detection
        assert event.guid != ""


def test_event_object_request_data() -> None:
    """Test EventObject request_data property."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="calendar123")
        data: dict[str, Any] = event.request_data
        assert "Event" in data
        assert "ClientState" in data
        assert data["Event"]["title"] == "New Event"
        assert "pGuid" in data["Event"]  # Note: camelCase in output
        assert data["Event"]["pGuid"] == "calendar123"
        assert "guid" in data["Event"]
        assert "Collection" in data["ClientState"]


def test_event_object_dt_to_list() -> None:
    """Test EventObject dt_to_list method."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="calendar123")
        dt = datetime(2023, 1, 1, 12, 30)
        result = event.dt_to_list(dt)
        assert result == ["20230101", 2023, 1, 1, 12, 30, 750]


def test_event_object_add_invitees() -> None:
    """Test EventObject add_invitees method."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="calendar123")
        event.add_invitees(["test@example.com", "user@example.com"])
        assert len(event.invitees) == 2
        assert f"{event.guid}:test@example.com" == event.invitees[0]
        assert f"{event.guid}:user@example.com" == event.invitees[1]


def test_event_object_dynamic_timezone() -> None:
    """Test that EventObject uses dynamic timezone detection based on user's locale."""
    # Test with different timezones to ensure dynamic behavior
    with patch(
        "pyicloud.services.calendar.get_localzone_name", return_value="Europe/London"
    ):
        event = EventObject(pguid="calendar123")
        assert event.tz == "Europe/London"

    with patch(
        "pyicloud.services.calendar.get_localzone_name", return_value="Asia/Tokyo"
    ):
        event = EventObject(pguid="calendar123")
        assert event.tz == "Asia/Tokyo"

    with patch(
        "pyicloud.services.calendar.get_localzone_name", return_value="America/New_York"
    ):
        event = EventObject(pguid="calendar123")
        assert event.tz == "America/New_York"


def test_calendar_object_initialization() -> None:
    """Test CalendarObject initialization and default values."""
    calendar = CalendarObject(title="My Calendar")
    assert calendar.title == "My Calendar"
    assert calendar.guid != ""
    assert calendar.color.startswith("#")


def test_calendar_object_request_data() -> None:
    """Test CalendarObject request_data property."""
    calendar = CalendarObject(title="My Calendar")
    data: dict[str, Any] = calendar.request_data
    assert "Collection" in data
    assert data["Collection"]["title"] == "My Calendar"
    assert "ClientState" in data
    assert "guid" in data["Collection"]
    assert "color" in data["Collection"]


def test_calendar_service_get_calendars() -> None:
    """Test CalendarService get_calendars method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Collection": [{"title": "Test Calendar"}]}
    mock_session.get.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        calendars = service.get_calendars()
        assert len(calendars) == 1
        assert calendars[0]["title"] == "Test Calendar"


def test_calendar_service_add_calendar() -> None:
    """Test CalendarService add_calendar method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        calendar = CalendarObject(title="New Calendar")
        response = service.add_calendar(calendar)
        assert response["status"] == "success"


def test_calendar_service_remove_calendar() -> None:
    """Test CalendarService remove_calendar method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response

    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        response = service.remove_calendar("calendar123")
        assert response["status"] == "success"


def test_calendar_service_get_events() -> None:
    """Test CalendarService get_events method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Event": [{"title": "Test Event"}]}
    mock_session.get.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        events = service.get_events()
        assert len(events) == 1
        assert events[0]["title"] == "Test Event"


def test_calendar_service_add_event() -> None:
    """Test CalendarService add_event method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.get_ctag = MagicMock(return_value="etag123")
        event = EventObject(pguid="calendar123", title="New Event")
        response = service.add_event(event)
        assert response["status"] == "success"


def test_calendar_service_remove_event() -> None:
    """Test CalendarService remove_event method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.get_ctag = MagicMock(return_value="etag123")

        event = EventObject(pguid="calendar123", title="New Event")
        response = service.remove_event(event)
        assert response["status"] == "success"


# =====================================
# Tests for NEW Features Added in PR
# =====================================


def test_constants_and_defaults() -> None:
    """Test all constant classes have expected values."""
    # Test DateFormats
    assert DateFormats.API_DATE == "%Y-%m-%d"
    assert DateFormats.APPLE_DATE == "%Y%m%d"

    # Test CalendarDefaults
    assert CalendarDefaults.TITLE == "Untitled"
    assert CalendarDefaults.SYMBOLIC_COLOR == "__custom__"
    assert CalendarDefaults.SUPPORTED_TYPE == "Event"
    assert CalendarDefaults.OBJECT_TYPE == "personal"
    assert CalendarDefaults.ORDER == 7
    assert CalendarDefaults.SHARE_TITLE == ""
    assert CalendarDefaults.SHARED_URL == ""
    assert CalendarDefaults.COLOR == ""

    # Test AlarmDefaults
    assert AlarmDefaults.MESSAGE_TYPE == "message"
    assert not AlarmDefaults.IS_LOCATION_BASED


def _service_with_mocks(mock_session: PyiCloudSession) -> CalendarService:
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        return CalendarService("https://example.com", mock_session, {"dsid": "12345"})


class _FixedDateTime(datetime):
    """Subclass datetime to control today() for tests."""

    fixed: datetime = datetime(2025, 2, 10)

    @classmethod
    def today(cls) -> "_FixedDateTime":  # type: ignore[override]
        return cls.fromtimestamp(cls.fixed.timestamp())


def test_default_params_feb_non_leap() -> None:
    """default_params should compute Feb (non-leap) as 1..28."""
    mock_session = MagicMock(spec=PyiCloudSession)
    service = _service_with_mocks(mock_session)

    # Freeze 'today' to 2025-02-10 (non-leap year)
    _FixedDateTime.fixed = datetime(2025, 2, 10)
    with (
        patch("pyicloud.services.calendar.datetime", _FixedDateTime),
        patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"),
    ):
        params = service.default_params
        assert params["startDate"] == "2025-02-01"
        assert params["endDate"] == "2025-02-28"


def test_default_params_feb_leap() -> None:
    """default_params should compute Feb (leap year) as 1..29."""
    mock_session = MagicMock(spec=PyiCloudSession)
    service = _service_with_mocks(mock_session)

    # Freeze 'today' to 2028-02-10 (leap year)
    _FixedDateTime.fixed = datetime(2028, 2, 10)
    with (
        patch("pyicloud.services.calendar.datetime", _FixedDateTime),
        patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"),
    ):
        params = service.default_params
        assert params["startDate"] == "2028-02-01"
        assert params["endDate"] == "2028-02-29"


def test_refresh_client_anchors_from_dt_month() -> None:
    """When only from_dt is provided, anchor to its month for the end bound."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Event": []}
    mock_session.get.return_value = mock_response
    service = _service_with_mocks(mock_session)

    from_dt = datetime(2025, 3, 15)
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service.refresh_client(from_dt=from_dt, to_dt=None)

    # Inspect params passed to GET
    _, kwargs = mock_session.get.call_args
    params = kwargs["params"]
    assert params["startDate"] == "2025-03-01"
    assert params["endDate"] == "2025-03-31"


def test_refresh_client_anchors_to_dt_month() -> None:
    """When only to_dt is provided, anchor to its month for the start bound."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Event": []}
    mock_session.get.return_value = mock_response
    service = _service_with_mocks(mock_session)

    to_dt = datetime(2025, 4, 20)
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service.refresh_client(from_dt=None, to_dt=to_dt)

    # Inspect params passed to GET
    _, kwargs = mock_session.get.call_args
    params = kwargs["params"]
    assert params["startDate"] == "2025-04-01"
    assert params["endDate"] == "2025-04-30"


def test_apple_date_format_dataclass() -> None:
    """Test AppleDateFormat dataclass functionality."""
    # Test from_datetime for start time
    dt = datetime(2023, 6, 15, 14, 30)
    apple_format = AppleDateFormat.from_datetime(dt, is_start=True)

    assert apple_format.date_string == "20230615"
    assert apple_format.year == 2023
    assert apple_format.month == 6
    assert apple_format.day == 15
    assert apple_format.hour == 14
    assert apple_format.minute == 30
    assert apple_format.minutes_from_midnight == 870  # 14*60 + 30

    # Test to_list conversion
    result_list = apple_format.to_list()
    expected = ["20230615", 2023, 6, 15, 14, 30, 870]
    assert result_list == expected

    # Test from_datetime for end time (different calculation)
    apple_format_end = AppleDateFormat.from_datetime(dt, is_start=False)
    assert (
        apple_format_end.minutes_from_midnight == 630
    )  # (24-14)*60 + (60-30) = 10*60 + 30


def test_calendar_object_uses_defaults() -> None:
    """Test CalendarObject uses constant defaults correctly."""
    calendar = CalendarObject()

    assert calendar.title == CalendarDefaults.TITLE
    assert calendar.symbolic_color == CalendarDefaults.SYMBOLIC_COLOR
    assert calendar.supported_type == CalendarDefaults.SUPPORTED_TYPE
    assert calendar.object_type == CalendarDefaults.OBJECT_TYPE
    assert calendar.order == CalendarDefaults.ORDER
    assert calendar.share_title == CalendarDefaults.SHARE_TITLE
    assert calendar.shared_url == CalendarDefaults.SHARED_URL
    # Color gets generated, so just check it's not empty
    assert calendar.color.startswith("#")


def test_event_object_validation() -> None:
    """Test EventObject validation logic."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        # Test empty pguid validation
        with pytest.raises(ValueError) as excinfo:
            EventObject(pguid="")
        assert "pguid cannot be empty" in str(excinfo.value)

        # Test empty pguid with whitespace
        with pytest.raises(ValueError) as excinfo:
            EventObject(pguid="   ")
        assert "pguid cannot be empty" in str(excinfo.value)

        # Test invalid date range (start after end)
        with pytest.raises(ValueError) as excinfo:
            EventObject(
                pguid="test-calendar",
                start_date=datetime(2023, 6, 15, 15, 0),
                end_date=datetime(2023, 6, 15, 14, 0),  # Earlier than start
            )
        assert "start_date" in str(excinfo.value) and "must be before end_date" in str(
            excinfo.value
        )

        # Test valid event creation
        event = EventObject(
            pguid="test-calendar",
            title="Valid Event",
            start_date=datetime(2023, 6, 15, 14, 0),
            end_date=datetime(2023, 6, 15, 15, 0),
        )
        assert event.pguid == "test-calendar"
        assert event.title == "Valid Event"
        assert event.duration == 60  # 1 hour in minutes


def test_event_object_alarm_functionality() -> None:
    """Test alarm creation and management."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="test-calendar", title="Alarm Test Event")

        # Test add_alarm_at_time
        alarm_guid = event.add_alarm_at_time()
        assert alarm_guid in event.alarms[0]  # Format: "eventGuid:alarmGuid"
        assert len(event.alarms) == 1
        assert event.alarms[0].startswith(event.guid)

        # Verify alarm metadata for "at time" alarm
        alarm_full_guid = event.alarms[0]
        assert alarm_full_guid in event._alarm_metadata
        metadata = event._alarm_metadata[alarm_full_guid]
        assert not metadata.before
        assert metadata.minutes == 0
        assert metadata.hours == 0

        # Test add_alarm_before with different time periods
        alarm_guid_5min = event.add_alarm_before(minutes=5)
        event.add_alarm_before(hours=1)  # Don't need to store this one
        alarm_guid_complex = event.add_alarm_before(days=1, hours=2, minutes=30)

        assert len(event.alarms) == 4  # 1 at-time + 3 before alarms

        # Verify "5 minutes before" alarm metadata
        alarm_5min_full = f"{event.guid}:{alarm_guid_5min}"
        metadata_5min = event._alarm_metadata[alarm_5min_full]
        assert metadata_5min.before
        assert metadata_5min.minutes == 5
        assert metadata_5min.hours == 0
        assert metadata_5min.days == 0

        # Verify "complex before" alarm metadata
        alarm_complex_full = f"{event.guid}:{alarm_guid_complex}"
        metadata_complex = event._alarm_metadata[alarm_complex_full]
        assert metadata_complex.before
        assert metadata_complex.days == 1
        assert metadata_complex.hours == 2
        assert metadata_complex.minutes == 30


def test_event_object_alarm_payload_structure() -> None:
    """Test alarm payload structure in request_data."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="test-calendar", title="Alarm Test Event")

        # Add alarms
        event.add_alarm_at_time()
        event.add_alarm_before(minutes=15)

        # Get request data
        request_data = event.request_data

        # Verify Alarm array structure
        assert "Alarm" in request_data
        assert len(request_data["Alarm"]) == 2

        # Check first alarm structure (at time)
        alarm1 = request_data["Alarm"][0]
        assert "guid" in alarm1
        assert "pGuid" in alarm1
        assert "messageType" in alarm1
        assert "isLocationBased" in alarm1
        assert "measurement" in alarm1

        assert alarm1["pGuid"] == event.guid  # Event GUID, not calendar GUID
        assert alarm1["messageType"] == AlarmDefaults.MESSAGE_TYPE
        assert alarm1["isLocationBased"] == AlarmDefaults.IS_LOCATION_BASED

        # Check measurement structure
        measurement1 = alarm1["measurement"]
        assert "before" in measurement1
        assert "minutes" in measurement1
        assert "hours" in measurement1
        assert "days" in measurement1

        # Verify Event.alarms field contains correct string format
        event_data = request_data["Event"]
        assert "alarms" in event_data
        assert len(event_data["alarms"]) == 2
        assert all(
            ":" in alarm for alarm in event_data["alarms"]
        )  # Format: "eventGuid:alarmGuid"


def test_event_object_invitee_payload_structure() -> None:
    """Test invitee payload structure in request_data."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(pguid="test-calendar", title="Invitee Test Event")

        # Add invitees
        event.add_invitees(["test@example.com", "user@example.com"])

        # Get request data
        request_data = event.request_data

        # Verify Invitee array structure
        assert "Invitee" in request_data
        assert len(request_data["Invitee"]) == 2

        # Check first invitee structure
        invitee1 = request_data["Invitee"][0]
        assert "guid" in invitee1
        assert "pGuid" in invitee1
        assert "role" in invitee1
        assert "isOrganizer" in invitee1
        assert "email" in invitee1
        assert "inviteeStatus" in invitee1
        assert "commonName" in invitee1
        assert "isMe" in invitee1  # Should be "isMe", not "isMyId"

        assert invitee1["pGuid"] == event.guid  # Event GUID, not calendar GUID
        assert invitee1["email"] == "test@example.com"
        assert not invitee1["isMe"]

        # Verify Event.invitees field contains correct string format
        event_data = request_data["Event"]
        assert "invitees" in event_data
        assert len(event_data["invitees"]) == 2
        assert event_data["invitees"][0] == f"{event.guid}:test@example.com"
        assert event_data["invitees"][1] == f"{event.guid}:user@example.com"


def test_calendar_service_guid_bug_fix() -> None:
    """Test that GUID vs Calendar GUID bug is fixed."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response

    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )

        # Mock get_ctag to verify it's called with calendar GUID, not event GUID
        def mock_get_ctag(guid):
            # This should be called with the calendar GUID (event.pguid)
            # NOT the event GUID (event.guid)
            assert guid == "calendar-guid-123"
            return "test-ctag"

        service.get_ctag = mock_get_ctag

        # Create event with different event GUID and calendar GUID
        event = EventObject(pguid="calendar-guid-123", title="Test Event")
        event.guid = "event-guid-456"  # Different from pguid

        # This should work - get_ctag should be called with calendar GUID
        response = service.add_event(event)
        assert response["status"] == "success"

        # For remove_event as well
        response = service.remove_event(event)
        assert response["status"] == "success"


def test_complete_payload_structure() -> None:
    """Test that generated payload matches Apple's expected JSON structure."""
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        event = EventObject(
            pguid="test-calendar-guid",
            title="Complete Test Event",
            start_date=datetime(2023, 6, 15, 14, 0),
            end_date=datetime(2023, 6, 15, 15, 0),
            location="Test Location",
            all_day=False,
        )

        # Add invitees and alarms for complete test
        event.add_invitees(["test@example.com"])
        event.add_alarm_before(minutes=15)

        request_data = event.request_data

        # Verify top-level structure
        expected_keys = ["Event", "Invitee", "Alarm", "ClientState"]
        for key in expected_keys:
            assert key in request_data, f"Missing top-level key: {key}"

        # Verify Event structure has camelCase fields
        event_data = request_data["Event"]
        camelcase_fields = [
            "pGuid",
            "startDate",
            "endDate",
            "localStartDate",
            "localEndDate",
            "createdDate",
            "lastModifiedDate",
            "extendedDetailsAreIncluded",
            "recurrenceException",
            "recurrenceMaster",
            "hasAttachments",
            "shouldShowJunkUIWhenAppropriate",
            "changeRecurring",
            "allDay",
        ]
        for field in camelcase_fields:
            assert field in event_data, f"Missing camelCase field: {field}"

        # Verify date fields are in Apple's 7-element format
        assert isinstance(event_data["startDate"], list)
        assert len(event_data["startDate"]) == 7
        assert event_data["startDate"][0] == "20230615"  # YYYYMMDD string
        assert event_data["startDate"][1] == 2023  # Year
        assert event_data["startDate"][2] == 6  # Month

        # Verify ClientState structure
        client_state = request_data["ClientState"]
        assert "Collection" in client_state
        assert len(client_state["Collection"]) == 1
        collection = client_state["Collection"][0]
        assert collection["guid"] == event.pguid  # Calendar GUID, not event GUID
        assert "ctag" in collection


def test_alarm_measurement_dataclass() -> None:
    """Test AlarmMeasurement dataclass."""
    # Test default values
    measurement = AlarmMeasurement()
    assert measurement.before
    assert measurement.weeks == 0
    assert measurement.days == 0
    assert measurement.hours == 0
    assert measurement.minutes == 0
    assert measurement.seconds == 0

    # Test custom values
    measurement = AlarmMeasurement(before=False, days=1, hours=2, minutes=30)
    assert not measurement.before
    assert measurement.days == 1
    assert measurement.hours == 2
    assert measurement.minutes == 30


def test_apple_alarm_dataclass() -> None:
    """Test AppleAlarm dataclass."""
    measurement = AlarmMeasurement(before=True, minutes=15)
    alarm = AppleAlarm(
        guid="event-guid:alarm-guid", pGuid="event-guid", measurement=measurement
    )

    assert alarm.guid == "event-guid:alarm-guid"
    assert alarm.pGuid == "event-guid"
    assert alarm.messageType == AlarmDefaults.MESSAGE_TYPE
    assert alarm.isLocationBased == AlarmDefaults.IS_LOCATION_BASED
    assert alarm.measurement.minutes == 15



"""Base service."""

from abc import ABC
from typing import Any

from pyicloud.session import PyiCloudSession


class BaseService(ABC):
    """The base iCloud service."""

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        self.__session: PyiCloudSession = session
        self.__params: dict[str, Any] = params
        self.__service_root: str = service_root

    @property
    def session(self) -> PyiCloudSession:
        """The session object."""
        return self.__session

    @property
    def params(self) -> dict[str, Any]:
        """The request parameters."""
        return self.__params

    @property
    def service_root(self) -> str:
        """The service root URL."""
        return self.__service_root



"""Find My iPhone service tests."""

# pylint: disable=protected-access

from datetime import datetime, timedelta
from unittest.mock import MagicMock, PropertyMock, call, patch

import pytest

from pyicloud import PyiCloudService
from pyicloud.exceptions import (
    PyiCloudAuthRequiredException,
    PyiCloudNoDevicesException,
    PyiCloudServiceUnavailable,
)
from pyicloud.services.findmyiphone import (
    AppleDevice,
    FindMyiPhoneServiceManager,
    _monitor_thread,
)
from tests.const.const_findmyiphone import FMI_FAMILY_WORKING


def test_devices(pyicloud_service_working: PyiCloudService) -> None:
    """Tests devices."""

    assert pyicloud_service_working.devices

    for device in pyicloud_service_working.devices:
        assert device["canWipeAfterLock"] is not None
        assert device["baUUID"] is not None
        assert device["wipeInProgress"] is not None
        assert device["lostModeEnabled"] is not None
        assert device["activationLocked"] is not None
        assert device["passcodeLength"] is not None
        assert device["deviceStatus"] is not None
        assert device["features"] is not None
        assert device["lowPowerMode"] is not None
        assert device["rawDeviceModel"] is not None
        assert device["id"] is not None
        assert device["isLocating"] is not None
        assert device["modelDisplayName"] is not None
        assert device["lostTimestamp"] is not None
        assert device["batteryLevel"] is not None
        assert device["locationEnabled"] is not None
        assert device["locFoundEnabled"] is not None
        assert device["fmlyShare"] is not None
        assert device["lostModeCapable"] is not None
        assert device["wipedTimestamp"] is None
        assert device["deviceDisplayName"] is not None
        assert device["audioChannels"] is not None
        assert device["locationCapable"] is not None
        assert device["batteryStatus"] is not None
        assert device["trackingInfo"] is None
        assert device["name"] is not None
        assert device["isMac"] is not None
        assert device["thisDevice"] is not None
        assert device["deviceClass"] is not None
        assert device["deviceModel"] is not None
        assert device["maxMsgChar"] is not None
        assert device["darkWake"] is not None
        assert device["remoteWipe"] is None

        assert device.data["canWipeAfterLock"] is not None
        assert device.data["baUUID"] is not None
        assert device.data["wipeInProgress"] is not None
        assert device.data["lostModeEnabled"] is not None
        assert device.data["activationLocked"] is not None
        assert device.data["passcodeLength"] is not None
        assert device.data["deviceStatus"] is not None
        assert device.data["features"] is not None
        assert device.data["lowPowerMode"] is not None
        assert device.data["rawDeviceModel"] is not None
        assert device.data["id"] is not None
        assert device.data["isLocating"] is not None
        assert device.data["modelDisplayName"] is not None
        assert device.data["lostTimestamp"] is not None
        assert device.data["batteryLevel"] is not None
        assert device.data["locationEnabled"] is not None
        assert device.data["locFoundEnabled"] is not None
        assert device.data["fmlyShare"] is not None
        assert device.data["lostModeCapable"] is not None
        assert device.data["wipedTimestamp"] is None
        assert device.data["deviceDisplayName"] is not None
        assert device.data["audioChannels"] is not None
        assert device.data["locationCapable"] is not None
        assert device.data["batteryStatus"] is not None
        assert device.data["trackingInfo"] is None
        assert device.data["name"] is not None
        assert device.data["isMac"] is not None
        assert device.data["thisDevice"] is not None
        assert device.data["deviceClass"] is not None
        assert device.data["deviceModel"] is not None
        assert device.data["maxMsgChar"] is not None
        assert device.data["darkWake"] is not None
        assert device.data["remoteWipe"] is None


def test_apple_device_properties(pyicloud_service_working: PyiCloudService) -> None:
    """Tests AppleDevice properties and methods."""
    device: AppleDevice = pyicloud_service_working.devices[0]

    # Test session property
    assert device.session is not None

    # Test location property
    location = device.location
    assert location is not None
    assert "latitude" in location
    assert "longitude" in location
    with patch(
        "pyicloud.services.findmyiphone.AppleDevice.location_available",
        new_callable=PropertyMock,
    ) as mock_location_available:
        mock_location_available.return_value = False
        assert device.location is None

    # Test status method
    status = device.status()
    assert "batteryLevel" in status
    assert "deviceDisplayName" in status
    assert "deviceStatus" in status
    assert "name" in status

    # Test status with additional fields
    additional_status = device.status(additional=["isMac", "deviceClass"])
    assert "isMac" in additional_status
    assert "deviceClass" in additional_status

    # Test data property
    assert device.data is not None
    assert "id" in device.data
    with (
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth"
        ) as mock_refresh,
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager.is_alive",
            new_callable=PropertyMock,
        ) as mock_is_alive,
    ):
        mock_is_alive.return_value = False
        assert "id" in device.data
        mock_refresh.assert_called_once()

    # Test model property
    assert device.model == device.data["deviceModel"]

    # Test device_type property
    assert device.device_type == device.data["deviceClass"]

    # Test __getitem__ method
    device_id = device.data["id"]
    with (
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth"
        ) as mock_refresh,
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager.is_alive",
            new_callable=PropertyMock,
        ) as mock_is_alive,
    ):
        mock_is_alive.side_effect = [True, False]
        assert device["id"] == device_id
        mock_refresh.assert_not_called()

        assert device["id"] == device_id
        assert mock_refresh.call_count == 1

    # Test __getattr__ method
    assert device.deviceDisplayName == device.data["deviceDisplayName"]
    display_name = device.data["deviceDisplayName"]
    with pytest.raises(AttributeError):
        _ = device.non_existent_attribute
    with (
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth"
        ) as mock_refresh,
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager.is_alive",
            new_callable=PropertyMock,
        ) as mock_is_alive,
    ):
        mock_is_alive.return_value = False
        assert device.deviceDisplayName == display_name
        mock_refresh.assert_called_once()

    # Test __str__ method
    assert str(device) == f"{device['deviceDisplayName']}: {device['name']}"

    # Test __repr__ method
    assert repr(device) == f"<AppleDevice({device})>"


def test_apple_device_actions(pyicloud_service_working: PyiCloudService) -> None:
    """Tests AppleDevice actions like play_sound, display_message, and lost_device."""
    device: AppleDevice = pyicloud_service_working.devices[0]

    # Mock session.post to avoid actual API calls
    with patch.object(device.session, "post") as mock_post:
        # Test play_sound
        with pytest.raises(PyiCloudServiceUnavailable):
            device.data["features"] = {
                "WIP": False,
                "MSG": False,
                "LOC": False,
                "SND": False,
            }
            device.play_sound(subject="Test Alert")

        device.data["features"] = {"WIP": True, "MSG": True, "LOC": True, "SND": True}

        device.play_sound(subject="Test Alert")
        mock_post.assert_called_with(
            device._sound_url,
            params=device._params,
            json={
                "device": device.data["id"],
                "subject": "Test Alert",
                "clientContext": {"fmly": True},
            },
        )

        # Test display_message
        with pytest.raises(PyiCloudServiceUnavailable):
            device.data["features"] = {
                "WIP": False,
                "MSG": False,
                "LOC": False,
                "SND": False,
            }
            device.display_message(subject="Test Message", message="Hello", sounds=True)

        device.data["features"] = {"WIP": True, "MSG": True, "LOC": True, "SND": True}

        device.display_message(subject="Test Message", message="Hello", sounds=True)
        mock_post.assert_called_with(
            device._message_url,
            params=device._params,
            json={
                "device": device.data["id"],
                "subject": "Test Message",
                "sound": True,
                "vibrate": False,
                "strobe": False,
                "userText": True,
                "text": "Hello",
            },
        )

        # Test lost_device
        with pytest.raises(PyiCloudServiceUnavailable):
            device.data["lostModeCapable"] = False

            device.lost_device(
                number="1234567890", text="Lost device message", newpasscode="1234"
            )

        device.data["features"] = {"WIP": True, "MSG": True, "LOC": True, "SND": True}
        device.data["lostModeCapable"] = True
        device.lost_device(
            number="1234567890", text="Lost device message", newpasscode="1234"
        )
        mock_post.assert_called_with(
            device._lost_url,
            params=device._params,
            json={
                "text": "Lost device message",
                "userText": True,
                "ownerNbr": "1234567890",
                "lostModeEnabled": True,
                "trackingEnabled": True,
                "device": device.data["id"],
                "passcode": "1234",
            },
        )


def test_findmyiphone_service_manager(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Tests FindMyiPhoneServiceManager methods."""
    manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices

    # Test refresh_client
    manager._refresh_client_with_reauth()
    assert len(manager) > 0

    # Test __getitem__
    with (
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth"
        ) as mock_refresh,
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager.is_alive",
            new_callable=PropertyMock,
        ) as mock_is_alive,
    ):
        mock_is_alive.side_effect = [True, False, True, True]
        device: AppleDevice = manager[0]
        assert isinstance(device, AppleDevice)
        mock_refresh.assert_not_called()
        assert mock_is_alive.call_count == 1

        device: AppleDevice = manager[0]
        assert isinstance(device, AppleDevice)
        assert mock_refresh.call_count == 1
        assert mock_is_alive.call_count == 2

        device: AppleDevice = manager[device.data["id"]]
        assert isinstance(device, AppleDevice)
        assert mock_refresh.call_count == 1
        assert mock_is_alive.call_count == 4

    # Test __str__ and __repr__
    assert str(manager) == repr(manager)

    # Test __iter__
    with (
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth"
        ) as mock_refresh,
        patch(
            "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager.is_alive",
            new_callable=PropertyMock,
        ) as mock_is_alive,
    ):
        mock_is_alive.side_effect = [True, True, False, False]

        devices: list[AppleDevice] = list(iter(manager))
        assert len(devices) == len(manager)
        mock_refresh.assert_not_called()
        assert mock_is_alive.call_count == 2

        devices: list[AppleDevice] = list(iter(manager))
        assert len(devices) == len(manager)
        assert mock_refresh.call_count == 2
        assert mock_is_alive.call_count == 4

    assert len(manager.devices) == len(devices)
    assert manager.user_info == FMI_FAMILY_WORKING["userInfo"]


def test_refresh_no_content(pyicloud_service_working: PyiCloudService) -> None:
    """Tests refresh_client handles no content response."""
    with patch(
        "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth",
        return_value=None,
    ):
        manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices
        manager._with_family = True

        with patch.object(manager.session, "post") as mock_post:
            mock_post.return_value.json.return_value = {}
            manager._refresh_client()
            assert mock_post.call_count == 1
            assert len(manager._devices) == 0
            mock_post.assert_called_with(
                url=manager._fmip_init_url,
                params=manager.params,
                json={
                    "clientContext": {
                        "appName": "iCloud Find (Web)",
                        "appVersion": "2.0",
                        "apiVersion": "3.0",
                        "deviceListVersion": 1,
                        "fmly": True,
                        "timezone": "US/Pacific",
                        "inactiveTime": 0,
                    }
                },
            )


def test_refresh_with_server_ctx(pyicloud_service_working: PyiCloudService) -> None:
    """Tests refresh_client handles serverContext in response."""
    with patch(
        "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth",
        return_value=None,
    ):
        manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices
        manager._with_family = True

        with patch.object(manager.session, "post") as mock_post:
            mock_post.return_value.json.return_value = {
                "serverContext": {
                    "theftLoss": {
                        "status": "OFF",
                    }
                },
                "content": [],
                "error": None,
            }
            manager._refresh_client()
            manager._refresh_client()
            assert mock_post.call_count == 2
            assert len(manager._devices) == 0
            mock_post.assert_has_calls(
                [
                    call(
                        url=manager._fmip_init_url,
                        params=manager.params,
                        json={
                            "clientContext": {
                                "appName": "iCloud Find (Web)",
                                "appVersion": "2.0",
                                "apiVersion": "3.0",
                                "deviceListVersion": 1,
                                "fmly": True,
                                "timezone": "US/Pacific",
                                "inactiveTime": 0,
                            }
                        },
                    ),
                    call().json(),
                    call(
                        url=manager._fmip_refresh_url,
                        params=manager.params,
                        json={
                            "clientContext": {
                                "appName": "iCloud Find (Web)",
                                "appVersion": "2.0",
                                "apiVersion": "3.0",
                                "deviceListVersion": 1,
                                "fmly": True,
                                "timezone": "US/Pacific",
                                "inactiveTime": 0,
                            },
                            "serverContext": {"theftLoss": None},
                        },
                    ),
                    call().json(),
                ]
            )


def test_get_erase_token_success(pyicloud_service_working: PyiCloudService) -> None:
    """Tests AppleDevice._get_erase_token returns token when available."""
    device: AppleDevice = pyicloud_service_working.devices[0]
    expected_token = "test_erase_token"

    with patch.object(device.session, "post") as mock_post:
        mock_post.return_value.json.return_value = {
            "tokens": {"mmeFMIPWebEraseDeviceToken": expected_token}
        }
        token: str = device._get_erase_token()
        assert token == expected_token
        mock_post.assert_called_with(
            url=device._erase_token_url,
            json={"dsWebAuthToken": device.session.data.get("session_token")},
        )


def test_get_erase_token_missing_tokens(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Tests AppleDevice._get_erase_token raises when tokens missing."""
    device: AppleDevice = pyicloud_service_working.devices[0]

    with patch.object(device.session, "post") as mock_post:
        mock_post.return_value.json.return_value = {}

        with pytest.raises(PyiCloudServiceUnavailable):
            device._get_erase_token()

        mock_post.assert_called_with(
            url=device._erase_token_url,
            json={"dsWebAuthToken": device.session.data.get("session_token")},
        )


def test_get_erase_token_missing_token_key(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Tests AppleDevice._get_erase_token raises when token key missing."""
    device: AppleDevice = pyicloud_service_working.devices[0]

    with (
        patch.object(device.session, "post") as mock_post,
        pytest.raises(PyiCloudServiceUnavailable),
    ):
        mock_post.return_value.json.return_value = {"tokens": {}}
        device._get_erase_token()


def test_erase_device_calls_post_with_correct_data(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Tests AppleDevice.erase_device calls session.post with correct data."""
    device: AppleDevice = pyicloud_service_working.devices[0]
    expected_token = "test_erase_token"

    with (
        patch.object(
            device, "_get_erase_token", return_value=expected_token
        ) as mock_get_token,
        patch.object(device.session, "post") as mock_post,
    ):
        device.erase_device(text="Erase this device", newpasscode="5678")
        mock_get_token.assert_called_once()
        mock_post.assert_called_with(
            device._erase_url,
            params=device._params,
            json={
                "authToken": expected_token,
                "text": "Erase this device",
                "device": device.data["id"],
                "passcode": "5678",
            },
        )


def test_erase_device_default_arguments(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Tests AppleDevice.erase_device with default arguments."""
    device: AppleDevice = pyicloud_service_working.devices[0]
    expected_token = "default_token"

    with (
        patch.object(
            device, "_get_erase_token", return_value=expected_token
        ) as mock_get_token,
        patch.object(device.session, "post") as mock_post,
    ):
        with pytest.raises(PyiCloudServiceUnavailable):
            device.data["features"] = {
                "WIP": False,
            }
            device.erase_device()

        device.data["features"] = {
            "WIP": True,
        }
        device.erase_device()
        mock_get_token.assert_called_once()
        mock_post.assert_called_with(
            device._erase_url,
            params=device._params,
            json={
                "authToken": expected_token,
                "text": "This device has been lost. Please call me.",
                "device": device.data["id"],
                "passcode": "",
            },
        )


def test_refresh_client_with_reauth_auth_required(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Test refresh_client_with_reauth handles PyiCloudAuthRequiredException and reauthenticates."""
    manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices

    # Patch _refresh_client to raise PyiCloudAuthRequiredException first, then succeed
    with (
        patch.object(
            manager,
            "_refresh_client",
            side_effect=[PyiCloudAuthRequiredException("", MagicMock()), None],
        ) as mock_refresh,
        patch.object(manager.session.service, "authenticate") as mock_authenticate,
        patch.object(manager, "_devices", {"dummy_id": "dummy_device"}),
        patch.object(manager, "_with_family", False),
    ):
        manager._refresh_client_with_reauth()
        mock_authenticate.assert_called_once_with(force_refresh=True)
        assert mock_refresh.call_count == 2
        mock_refresh.assert_has_calls([call(locate=True), call(locate=True)])


def test_refresh_client_with_reauth_failed(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Test refresh_client_with_reauth handles PyiCloudAuthRequiredException and reauthenticates."""
    manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices

    # Patch _refresh_client to raise PyiCloudAuthRequiredException first, then succeed
    with (
        patch.object(
            manager,
            "_refresh_client",
            side_effect=[
                PyiCloudAuthRequiredException("", MagicMock()),
                PyiCloudAuthRequiredException("", MagicMock()),
            ],
        ) as mock_refresh,
        patch.object(manager.session.service, "authenticate") as mock_authenticate,
        patch.object(manager, "_devices", {"dummy_id": "dummy_device"}),
        patch.object(manager, "_with_family", False),
    ):
        with pytest.raises(PyiCloudAuthRequiredException):
            manager._refresh_client_with_reauth()
        mock_authenticate.assert_called_once_with(force_refresh=True)
        assert mock_refresh.call_count == 2
        mock_refresh.assert_has_calls([call(locate=True), call(locate=True)])


def test_refresh_client_with_reauth_with_locate(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Test refresh_client_with_reauth calls _refresh_client with locate=True."""
    manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices
    manager._with_family = True

    with (
        patch.object(manager, "_refresh_client") as mock_refresh,
        patch.object(manager, "_devices", {"dummy_id": "dummy_device"}),
    ):
        manager._refresh_client_with_reauth()
        # Should call _refresh_client once: with locate=True
        assert mock_refresh.call_count == 1
        mock_refresh.assert_any_call(locate=True)


def test_refresh_client_with_reauth_with_loading_to_done(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Test refresh_client_with_reauth calls _refresh_client if the members are loading."""
    with patch(
        "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth",
        return_value=None,
    ):
        manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices
    manager._with_family = True

    with (
        patch("time.sleep", return_value=None),
        patch.object(manager, "_refresh_client") as mock_refresh,
        patch.object(manager, "_user_info") as mock_user_info,
        patch.object(manager, "_devices", {"dummy_id": "dummy_device"}),
    ):
        mock_user_info.__getitem__.return_value = True

        mock_user_info.get.side_effect = [
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "LOADING",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "DONE",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
        ]
        manager._refresh_client_with_reauth()
        assert mock_refresh.call_count == 3
        mock_refresh.assert_any_call(locate=True)


def test_refresh_client_with_reauth_with_loading_no_complete(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Test refresh_client_with_reauth calls _refresh_client if the members are loading."""
    with patch(
        "pyicloud.services.findmyiphone.FindMyiPhoneServiceManager._refresh_client_with_reauth",
        return_value=None,
    ):
        manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices
    manager._with_family = True

    with (
        patch("time.sleep", return_value=None),
        patch.object(manager, "_refresh_client") as mock_refresh,
        patch.object(manager, "_user_info") as mock_user_info,
        patch.object(manager, "_devices", {"dummy_id": "dummy_device"}),
    ):
        mock_user_info.__getitem__.return_value = True

        mock_user_info.get.side_effect = [
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "LOADING",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "DONE",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
            True,
            {
                "member1": {
                    "firstName": "Member1",
                    "lastName": "One",
                    "appleId": "member1@example.com",
                    "deviceFetchStatus": "LOADING",
                },
                "member2": {
                    "firstName": "Member2",
                    "lastName": "Two",
                    "appleId": "member2@example.com",
                    "deviceFetchStatus": "DONE",
                },
            },
        ]
        manager._refresh_client_with_reauth()
        assert mock_refresh.call_count == 6
        mock_refresh.assert_called_with(locate=True)


def test_refresh_client_with_reauth_no_devices_raises(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Test refresh_client_with_reauth raises PyiCloudNoDevicesException when no devices."""
    manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices

    with (
        patch.object(manager, "_refresh_client"),
        patch.object(manager, "_devices", {}),
    ):
        with pytest.raises(PyiCloudNoDevicesException):
            manager._refresh_client_with_reauth()


def test_monitor_thread_calls_func_at_interval() -> None:
    """Test _monitor_thread calls function at specified interval."""

    mock_func = MagicMock()
    interval = 0.2

    with (
        patch("pyicloud.services.findmyiphone.datetime") as mock_datetime,
    ):
        mock_event = MagicMock()
        mock_event.wait.side_effect = [False, False, True]
        # Mock datetime.now() to simulate time progression
        base_time = datetime(2023, 1, 1, 12, 0, 0)
        mock_datetime.now.side_effect = [
            base_time,  # Initial next_event calculation
            base_time + timedelta(seconds=0.1),  # First loop check (not ready)
            base_time + timedelta(seconds=0.3),  # Second loop check (ready)
            base_time + timedelta(seconds=0.3),  # New next_event calculation
        ]
        mock_datetime.side_effect = datetime

        _monitor_thread(interval, mock_func, mock_event, locate=True)

        # Should call func once when interval has passed
        mock_func.assert_called_once_with(True)


def test_monitor_thread_passes_locate_parameter() -> None:
    """Test _monitor_thread passes locate parameter to function."""

    mock_func = MagicMock()

    with (
        patch("pyicloud.services.findmyiphone.datetime") as mock_datetime,
    ):
        mock_event = MagicMock()
        mock_event.wait.side_effect = [False, True]

        base_time = datetime(2023, 1, 1, 12, 0, 0)
        mock_datetime.now.side_effect = [
            base_time,  # Initial next_event
            base_time + timedelta(seconds=1.0),  # Loop check (ready)
            base_time + timedelta(seconds=1.0),  # New next_event
        ]
        mock_datetime.side_effect = datetime

        _monitor_thread(0.5, mock_func, mock_event, locate=False)

        mock_func.assert_called_once_with(False)


def test_monitor_thread_handles_exception() -> None:
    """Test _monitor_thread handles the function raising an exception."""

    mock_func = MagicMock()
    mock_func.side_effect = Exception("Test Exception")

    with (
        patch("pyicloud.services.findmyiphone.datetime") as mock_datetime,
    ):
        mock_event = MagicMock()
        mock_event.wait.side_effect = [False, True]

        base_time = datetime(2023, 1, 1, 12, 0, 0)
        mock_datetime.now.side_effect = [
            base_time,  # Initial next_event
            base_time + timedelta(seconds=1.0),  # Loop check (ready)
            base_time + timedelta(seconds=1.0),  # New next_event
        ]
        mock_datetime.side_effect = datetime

        _monitor_thread(0.5, mock_func, mock_event, locate=False)

        mock_func.assert_called_once_with(False)


def test_monitor_thread_multiple_intervals() -> None:
    """Test _monitor_thread calls function multiple times across intervals."""

    mock_func = MagicMock()
    interval = 0.1

    with (
        patch("pyicloud.services.findmyiphone.datetime") as mock_datetime,
    ):
        # Main thread alive for multiple iterations
        mock_event = MagicMock()
        mock_event.wait.side_effect = [False, False, False, True]
        base_time = datetime(2023, 1, 1, 12, 0, 0)
        mock_datetime.now.side_effect = [
            base_time,  # Initial next_event
            base_time + timedelta(seconds=0.15),  # First ready
            base_time + timedelta(seconds=0.15),  # New next_event after first call
            base_time + timedelta(seconds=0.25),  # Next not ready (before interval)
            base_time + timedelta(seconds=0.26),  # Second ready
            base_time + timedelta(seconds=0.30),  # New next_event after second call
        ]
        mock_datetime.side_effect = datetime

        _monitor_thread(interval, mock_func, mock_event, locate=True)

        # Should call func twice
        assert mock_func.call_count == 2
        mock_func.assert_has_calls([call(True), call(True)])



#!/usr/bin/env python
"""End to End System test"""

import argparse
import http.client
import json
import logging
import sys
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, List, Optional
from unittest.mock import patch

import click
from fido2.hid import CtapHidDevice
from requests import Response

from pyicloud import PyiCloudService
from pyicloud.exceptions import PyiCloudServiceUnavailable
from pyicloud.services.calendar import CalendarObject, CalendarService
from pyicloud.services.photos import BasePhotoAlbum, PhotoAlbum, PhotoAsset
from pyicloud.ssl_context import configurable_ssl_verification

END_LIST: str = "End List\n"
MAX_DISPLAY: int = 10

# Set to FALSE to disable SSL verification to use tools like charles, mitmproxy, fiddler, or similiar tools to debug the data sent on the wire.
# Can also use command-line argument --disable-ssl-verify
# This uses code taken from:
# - https://stackoverflow.com/questions/15445981/how-do-i-disable-the-security-certificate-check-in-python-requests
# - https://stackoverflow.com/questions/16337511/log-all-requests-from-the-python-requests-module
ENABLE_SSL_VERIFICATION: bool = True

# Set the log level for HTTP commands
HTTP_LOG_LEVEL: int = logging.ERROR

# Set the log level for other commands
OTHER_LOG_LEVEL: int = logging.ERROR

# HTTPConnection parameters
HTTPCONNECTION_DEBUG_INFO: bool = False
HTTP_PROXY: Optional[str] = None
HTTPS_PROXY: Optional[str] = None

# Set where you'd like the COOKIES to be stored. Can also use command-line argument --cookie-dir
COOKIE_DIR: str = ""  # location to store session information

# Other configurable variables
APPLE_USERNAME: str = ""
APPLE_PASSWORD: str = ""
CHINA: bool = False


def parse_args() -> None:
    """Parse command line arguments"""
    global ENABLE_SSL_VERIFICATION, COOKIE_DIR, APPLE_PASSWORD, APPLE_USERNAME, CHINA, HTTP_PROXY, HTTPS_PROXY  # pylint: disable=global-statement
    parser = argparse.ArgumentParser(description="End to End Test of Services")

    parser.add_argument(
        "--username",
        action="store",
        dest="username",
        default="",
        help="Apple ID to Use",
    )
    parser.add_argument(
        "--password",
        action="store",
        dest="password",
        default="",
        help=(
            "Apple ID Password to Use; if unspecified, password will be "
            "fetched from the system keyring."
        ),
    )
    parser.add_argument(
        "--cookie-dir",
        action="store",
        dest="cookie_directory",
        default="",
        help="Directory to store session information and cookies",
    )
    parser.add_argument(
        "--china-mainland",
        action="store_true",
        dest="china_mainland",
        default=False,
        help="If the country/region setting of the Apple ID is China mainland",
    )
    parser.add_argument(
        "--disable-ssl-verify",
        action="store_true",
        dest="disable_ssl",
        default=False,
        help="Disable SSL verification",
    )

    parser.add_argument(
        "--http-proxy",
        type=str,
        help="Use HTTP proxy for requests",
    )

    parser.add_argument(
        "--https-proxy",
        type=str,
        help="Use HTTPS proxy for requests",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args: argparse.Namespace = parser.parse_args()

    if not args.username:
        parser.error("Both --username is required")
    else:
        APPLE_USERNAME = args.username
        APPLE_PASSWORD = args.password or ""

    if args.cookie_directory:
        COOKIE_DIR = args.cookie_directory

    if args.china_mainland:
        CHINA = args.china_mainland

    if args.disable_ssl or not ENABLE_SSL_VERIFICATION:
        ENABLE_SSL_VERIFICATION = False
        print("=" * 80)
        print("⚠️  SECURITY WARNING: SSL VERIFICATION DISABLED ⚠️")
        print("This is insecure and should ONLY be used for debugging!")
        print("Your credentials and data may be exposed to attackers.")
        print("=" * 80)
        print()

    if args.http_proxy:
        HTTP_PROXY = args.http_proxy
    if args.https_proxy:
        HTTPS_PROXY = args.https_proxy


def httpclient_logging_patch(level=HTTP_LOG_LEVEL) -> None:
    """Enable HTTPConnection debug logging to the logging framework"""
    httpclient_logger: logging.Logger = logging.getLogger("http.client")
    httpclient_logger.setLevel(level)

    def httpclient_log(*args) -> None:
        httpclient_logger.log(level, " ".join(map(str, args)))

    # mask the print() built-in in the http.client module to use
    # logging instead
    patch("http.client.print", httpclient_log).start()

    # enable debugging
    if HTTPCONNECTION_DEBUG_INFO:
        http.client.HTTPConnection.debuglevel = 1
    else:
        http.client.HTTPConnection.debuglevel = 0


def handle_2fa(api: PyiCloudService) -> None:
    """Handle two-factor authentication"""
    security_key_names: Optional[List[str]] = api.security_key_names

    if security_key_names:
        print(
            f"Security key confirmation is required. "
            f"Please plug in one of the following keys: {', '.join(security_key_names)}"
        )

        fido2_devices: List[CtapHidDevice] = api.fido2_devices

        if not fido2_devices:
            print("No FIDO2 devices detected. Connect a security key and try again.")
            sys.exit(1)

        print("Available FIDO2 devices:")

        for idx, dev in enumerate(fido2_devices, start=1):
            print(f"{idx}: {dev}")

        choice = click.prompt(
            "Select a FIDO2 device by number",
            type=click.IntRange(1, len(fido2_devices)),
            default=1,
        )
        selected_device: CtapHidDevice = fido2_devices[choice - 1]

        print("Please confirm the action using the security key")

        api.confirm_security_key(selected_device)

    else:
        print("Two-factor authentication required.")
        code: str = input(
            "Enter the code you received of one of your approved devices: "
        )
        result: bool = api.validate_2fa_code(code)
        print(f"Code validation result: {result}")

        if not result:
            print("Failed to verify security code")
            sys.exit(1)

    if not api.is_trusted_session:
        print("Session is not trusted. Requesting trust...")
        result = api.trust_session()
        print(f"Session trust result: {result}")

        if not result:
            print(
                "Failed to request trust. You will likely be prompted for confirmation again in the coming weeks"
            )


def handle_2sa(api: PyiCloudService) -> None:
    """Handle two-step authentication"""
    print("Two-step authentication required. Your trusted devices are:")

    trusted_devices: List[dict[str, Any]] = api.trusted_devices
    if not trusted_devices:
        print("No trusted devices are available for 2-step verification.")
        sys.exit(1)
    for i, device in enumerate(trusted_devices):
        print(
            "  %s: %s"
            % (i, device.get("deviceName", "SMS to %s" % device.get("phoneNumber")))
        )

    device_index: int = click.prompt(
        "Which device would you like to use?",
        type=click.IntRange(0, len(trusted_devices) - 1),
        default=0,
    )
    device: dict[str, Any] = trusted_devices[device_index]
    if not api.send_verification_code(device):
        print("Failed to send verification code")
        sys.exit(1)

    code = click.prompt("Please enter validation code")
    if not api.validate_verification_code(device, code):
        print("Failed to verify verification code")
        sys.exit(1)


def get_api() -> PyiCloudService:
    """Get authenticated PyiCloudService instance"""
    api = PyiCloudService(
        apple_id=APPLE_USERNAME,
        password=APPLE_PASSWORD,
        china_mainland=CHINA,
        cookie_directory=COOKIE_DIR,
    )

    if api.requires_2fa:
        handle_2fa(api)

    elif api.requires_2sa:
        handle_2sa(api)

    return api


def display_devices(api: PyiCloudService) -> None:
    """Display device info"""

    print(f"List of devices ({len(api.devices)}):")
    for idx, device in enumerate(api.devices):
        print(f"\t{idx}: {device}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    print("First device:")
    print(f"\t Name: {api.iphone}")
    print(f"\t Location: {json.dumps(api.iphone.location, indent=4)}\n")


def display_calendars(api: PyiCloudService) -> None:
    """Display calendar info"""

    calendar_service: CalendarService = api.calendar
    calendars: list[CalendarObject] = calendar_service.get_calendars(as_objs=True)
    print(f"List of calendars ({len(calendars)}):")
    for idx, calendar in enumerate(calendars):
        print(f"\t{idx}: {calendar.title}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    if not calendars:
        return

    # Get recent events from API
    try:
        recent_events = calendar_service.get_events(
            from_dt=datetime.now() - timedelta(days=7),
            to_dt=datetime.now() + timedelta(days=7),
            as_objs=True,
        )
        print(f"Recent events (±7 days): {len(recent_events)} events")
        for idx, event in enumerate(recent_events):
            if hasattr(event, "title") and hasattr(event, "start_date"):
                print(f"\t{idx}: {event.title} at {event.start_date}")
                if idx >= MAX_DISPLAY - 1:
                    break
        print(END_LIST)
    except Exception as e:
        print(f"Could not retrieve events: {e}\n")


def display_contacts(api: PyiCloudService) -> None:
    """Display contacts info"""

    contacts: List[dict[str, Any]] | None = api.contacts.all
    if contacts:
        print(f"List of contacts ({len(contacts)}):")
        for idx, contact in enumerate(contacts):
            print(
                f"\t{idx}: {contact.get('firstName') or contact.get('lastName') or contact.get('companyName')}"
            )
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)
    else:
        print("No contacts found\n")


def display_drive(api: PyiCloudService) -> None:
    """Display drive info"""

    drive_files: list[str] = api.drive.dir()
    print(f"List of files in iCloud Drive root ({len(drive_files)}):")
    for idx, filename in enumerate(drive_files):
        print(f"\t{idx}: {filename} ({api.drive[filename].type})")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_files(api: PyiCloudService) -> None:
    """Display files info"""

    try:
        files: list[str] = api.files.dir()
        print(f"List of files in iCloud files root ({len(files)}):")
        for idx, filename in enumerate(files):
            print(f"\t{idx}: {filename} ({api.files[filename].type})")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)
    except PyiCloudServiceUnavailable as error:
        print(f"Files service not available: {error}\n")


def display_photos(api: PyiCloudService) -> None:
    """Display photo info"""

    print(f"List of photo albums ({len(api.photos.albums)}):")
    for idx, album in enumerate(api.photos.albums):
        print(f"\t{idx}: {album}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    print(f"List of ALL PHOTOS ({len(api.photos.all)}):")
    for idx, photo in enumerate(api.photos.all):
        print(f"\t{idx}: {photo.filename} ({photo.item_type})")
        if idx >= MAX_DISPLAY - 1:
            data: bytes | None = photo.download()
            if data:
                print(f"\t\tDownloaded {len(data)} bytes")
            else:
                print("\t\tDownload failed")
            break
    print(END_LIST)


def display_videos(api: PyiCloudService) -> None:
    """Display video info"""
    if "Videos" in api.photos.albums:
        print(f"List of Videos ({len(api.photos.albums['Videos'])}):")
        for idx, photo in enumerate(api.photos.albums["Videos"]):
            print(f"\t{idx}: {photo.filename} ({photo.item_type})")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)
    else:
        print("No 'Videos' album found")


def display_shared_photos(api: PyiCloudService) -> None:
    """Display shared photo info"""

    selected_album: BasePhotoAlbum | None = next(iter(api.photos.shared_streams), None)
    print(f"List of Shared Albums ({len(api.photos.shared_streams)}):")
    for idx, album in enumerate(api.photos.shared_streams):
        print(f"\t{idx}: {album.name} ({len(album)} photos)")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    if selected_album and api.photos.shared_streams:
        print(f"List of Shared Photos [{selected_album.name}] ({len(selected_album)}):")
        for idx, photo in enumerate(selected_album):
            print(f"\t{idx}: {photo.filename} ({photo.item_type})")

            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def display_account(api: PyiCloudService) -> None:
    """Display account info"""

    print(f"Account name: {api.account_name}")
    print(f"Account plan: {json.dumps(api.account.summary_plan, indent=4)}")
    print(f"List of Family Member ({len(api.account.family)}):")
    for idx, member in enumerate(api.account.family):
        print(f"\t{idx}: {member}")
        try:
            photo: Response = member.get_photo()
            print(f"\t\tPhoto: {photo}")
            print(f"\t\tPhoto type: {photo.headers['Content-Type']}")
            print(f"\t\tPhoto size: {photo.headers['Content-Length']}")
        except Exception as e:
            print(f"\t\tPhoto: Error retrieving user photo: {e}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_hidemyemail(api: PyiCloudService) -> None:
    """Display Hide My Email info"""

    print(f"List of Hide My Email ({len(api.hidemyemail)}):")
    for idx, email in enumerate(api.hidemyemail):
        print(
            f"\t{idx}: {email['hme']} ({email['domain']}) Active = {email['isActive']}"
        )
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def album_management(api: PyiCloudService) -> None:
    """Test album management functions"""

    album_name = "Test Album from API"
    print(f"Creating album '{album_name}'...")
    album: PhotoAlbum | None = api.photos.create_album(album_name)
    print(f"Album created: {album}")
    if album is None:
        print("Album creation failed.")
        return

    print(f"Album '{album_name}' created successfully.")
    album.name = "Renamed Album"
    print(f"Album renamed to '{album.name}'")

    sample_photo: Path = Path(__file__).with_name("sample.jpg")
    if sample_photo.exists():
        photo: PhotoAsset | None = album.upload(str(sample_photo))
        if photo:
            print(f"Photo uploaded successfully: {photo.filename} ({photo.item_type})")
            if photo.delete():
                print("Photo deleted successfully.")
        else:
            print("Photo upload failed.")
    else:
        print(f"Skipping upload: sample photo not found at {sample_photo}")

    print(f"Deleting album '{album.name}'...")
    if album.delete():
        print("Album deleted.")
    else:
        print("Album deletion failed.")


def setup() -> None:
    """Setup"""
    parse_args()

    # Enable general debug logging
    logging.basicConfig(level=OTHER_LOG_LEVEL)

    # Enable httpclient logging
    httpclient_logging_patch()


def main() -> None:
    """main function"""
    with configurable_ssl_verification(
        ENABLE_SSL_VERIFICATION,
        HTTP_PROXY,
        HTTPS_PROXY,
    ):
        api: PyiCloudService = get_api()

        display_account(api)
        display_devices(api)
        display_hidemyemail(api)
        try:
            display_calendars(api)
        except PyiCloudServiceUnavailable as error:
            print(f"Calendar service not available: {error}\n")
        display_files(api)
        display_contacts(api)
        display_drive(api)
        display_photos(api)
        display_videos(api)
        display_shared_photos(api)
        album_management(api)


if __name__ == "__main__":
    setup()
    main()



"""Hide my email service."""

from typing import Any, Generator, Optional

from requests import Response

from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession


class HideMyEmailService(BaseService):
    """
    The 'Hide My Email' iCloud service connects to iCloud and manages email aliases.

    This service allows users to:
    - Generate new email aliases
    - Reserve specific aliases
    - List all existing aliases
    - Get alias details by ID
    - Update alias metadata (label, note)
    - Delete aliases
    - Deactivate aliases
    - Reactivate aliases
    """

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        super().__init__(service_root, session, params)

        # Define v1 endpoints
        self._v1_endpoint: str = f"{service_root}/v1/hme"
        self._generate_endpoint: str = f"{self._v1_endpoint}/generate"
        self._reserve_endpoint: str = f"{self._v1_endpoint}/reserve"
        self._update_metadata_endpoint: str = f"{self._v1_endpoint}/updateMetaData"
        self._delete_endpoint: str = f"{self._v1_endpoint}/delete"
        self._deactivate_endpoint: str = f"{self._v1_endpoint}/deactivate"
        self._reactivate_endpoint: str = f"{self._v1_endpoint}/reactivate"

        # Define v2 endpoints
        self._v2_endpoint: str = f"{service_root}/v2/hme"
        self._list_endpoint: str = f"{self._v2_endpoint}/list"
        self._get_endpoint: str = f"{self._v2_endpoint}/get"

    def generate(self) -> Optional[str]:
        """
        Generate a new email alias.

        Returns:
            The generated email address string or None if generation failed.
        """
        req: Response = self.session.post(self._generate_endpoint, params=self.params)
        response: dict[str, dict[str, str]] = req.json()
        result: Optional[dict[str, str]] = response.get("result")
        if result:
            return result.get("hme")
        return None

    def reserve(self, email: str, label: str, note="Generated") -> dict[str, Any]:
        """
        Reserve an alias for emails.

        Args:
            email: The email alias to reserve.
            label: A label for the email alias.
            note: An optional note for the email alias.

        Returns:
            The API's result containing details about the reserved alias.
        """
        req: Response = self.session.post(
            self._reserve_endpoint,
            params=self.params,
            json={
                "hme": email,
                "label": label,
                "note": note,
            },
        )
        response = req.json()
        return response.get("result", {})

    def __len__(self) -> int:
        """
        Get the number of emails
        """
        req: Response = self.session.get(self._list_endpoint, params=self.params)
        response: dict[str, dict[str, str]] = req.json()
        result: Optional[dict[str, str]] = response.get("result")
        if result:
            return len(result.get("hmeEmails", []))
        return 0

    def __iter__(self) -> Generator[Any, Any, None]:
        """
        Iterate over the list of emails
        """
        req: Response = self.session.get(self._list_endpoint, params=self.params)
        response: dict[str, dict[str, str]] = req.json()
        result: Optional[dict[str, str]] = response.get("result")
        if result:
            yield from result.get("hmeEmails", [])

    def __getitem__(self, anonymous_id: str) -> dict[str, Any]:
        """
        Get alias email details by anonymous_id.

        Args:
            anonymous_id: The unique identifier for the alias.

        Returns:
            A dictionary containing details about the alias.
        """
        req: Response = self.session.post(
            self._get_endpoint,
            params=self.params,
            json={"anonymousId": anonymous_id},
        )
        response = req.json()
        return response.get("result", {})

    def update_metadata(
        self, anonymous_id: str, label: str, note: Optional[str] = None
    ) -> dict[str, Any]:
        """
        Update metadata for an alias email.

        Args:
            anonymous_id: The unique identifier for the alias.
            label: The new label for the alias.
            note: The new note for the alias (optional).

        Returns:
            A dictionary containing the API response.
        """
        payload: dict[str, str] = {
            "anonymousId": anonymous_id,
            "label": label,
        }
        if note is not None:
            payload["note"] = note

        req: Response = self.session.post(
            self._update_metadata_endpoint,
            params=self.params,
            json=payload,
        )
        response = req.json()
        return response.get("result", {})

    def delete(self, anonymous_id: str) -> dict[str, Any]:
        """
        Delete an alias email.

        Args:
            anonymous_id: The unique identifier for the alias to delete.

        Returns:
            A dictionary containing the API response.
        """
        req: Response = self.session.post(
            self._delete_endpoint,
            params=self.params,
            json={"anonymousId": anonymous_id},
        )
        response = req.json()
        return response.get("result", {})

    def deactivate(self, anonymous_id: str) -> dict[str, Any]:
        """
        Deactivate an alias email.

        Deactivating an alias means emails sent to it will no longer be forwarded,
        but the alias remains in your list and can be reactivated later.

        Args:
            anonymous_id: The unique identifier for the alias to deactivate.

        Returns:
            A dictionary containing the API response.
        """
        req: Response = self.session.post(
            self._deactivate_endpoint,
            params=self.params,
            json={"anonymousId": anonymous_id},
        )
        response = req.json()
        return response.get("result", {})

    def reactivate(self, anonymous_id: str) -> dict[str, Any]:
        """
        Reactivate a previously deactivated alias email.

        Reactivating an alias means emails sent to it will be forwarded again
        to your primary inbox.

        Args:
            anonymous_id: The unique identifier for the alias to reactivate.

        Returns:
            A dictionary containing the API response.
        """
        req: Response = self.session.post(
            self._reactivate_endpoint,
            params=self.params,
            json={"anonymousId": anonymous_id},
        )
        response = req.json()
        return response.get("result", {})



"""Library tests."""

# pylint: disable=protected-access

import json
from io import BytesIO
from typing import Any, Optional
from unittest.mock import MagicMock

from requests import Response

from pyicloud import base
from tests.const import (
    ACCOUNT_DEVICES_WORKING,
    ACCOUNT_FAMILY_WORKING,
    ACCOUNT_STORAGE_WORKING,
    AUTH_OK,
    AUTHENTICATED_USER,
    DRIVE_FILE_DOWNLOAD_WORKING,
    DRIVE_FOLDER_WORKING,
    DRIVE_ROOT_INVALID,
    DRIVE_ROOT_WORKING,
    DRIVE_SUBFOLDER_WORKING,
    DRIVE_TRASH_DELETE_FOREVER_WORKING,
    DRIVE_TRASH_RECOVER_WORKING,
    DRIVE_TRASH_WORKING,
    FMI_FAMILY_WORKING,
    LOGIN_2FA,
    LOGIN_WORKING,
    REQUIRES_2FA_TOKEN,
    REQUIRES_2FA_USER,
    TRUSTED_DEVICE_1,
    TRUSTED_DEVICES,
    VALID_2FA_CODE,
    VALID_COOKIE,
    VALID_TOKEN,
    VALID_TOKENS,
    VALID_USERS,
    VERIFICATION_CODE_KO,
    VERIFICATION_CODE_OK,
)


class ResponseMock(Response):
    """Mocked Response."""

    def __init__(self, result, status_code=200, **kwargs) -> None:
        """Set up response mock."""
        Response.__init__(self)
        self.result = result
        self.status_code = status_code
        self.raw = kwargs.get("raw")
        self.headers = kwargs.get("headers", {})

    @property
    def text(self) -> str:
        """Return text."""
        return json.dumps(self.result)


class PyiCloudSessionMock(base.PyiCloudSession):
    """Mocked PyiCloudSession."""

    def _request(self, method, url, **kwargs) -> ResponseMock:
        """Make the request."""
        params = kwargs.get("params")
        headers = kwargs.get("headers")
        data = kwargs.get("json")
        if not data:
            data = json.loads(kwargs.get("data", "{}")) if kwargs.get("data") else {}

        if self.service._setup_endpoint in url:
            if resp := self._handle_setup_endpoint(url, method, data, headers):
                return resp

        if self.service._auth_endpoint in url:
            if resp := self._handle_auth_endpoint(url, method, data):
                return resp

        if resp := self._handle_other_endpoints(url, method, data, params):
            return resp

        raise ValueError("No valid response")

    def _handle_other_endpoints(
        self, url, method, data, params
    ) -> Optional[ResponseMock]:
        """Handle other endpoints."""
        if "device/getDevices" in url and method == "GET":
            return ResponseMock(ACCOUNT_DEVICES_WORKING)
        if "family/getFamilyDetails" in url and method == "GET":
            return ResponseMock(ACCOUNT_FAMILY_WORKING)
        if "setup/ws/1/storageUsageInfo" in url and method == "POST":
            return ResponseMock(ACCOUNT_STORAGE_WORKING)

        resp: Optional[ResponseMock] = None

        resp = self._handle_drive_endpoints_post(url, method, data)
        if resp:
            return resp

        resp = self._handle_drive_endpoints_get(url, method, params)
        if resp:
            return resp

        if "fmi" in url and method == "POST":
            return ResponseMock(FMI_FAMILY_WORKING)

    def _handle_drive_endpoints_post(self, url, method, data) -> Optional[ResponseMock]:
        """Handle drive endpoints post requests."""
        if "retrieveItemDetailsInFolders" in url and method == "POST":
            if resp := self._handle_drive_retrieve(data):
                return resp

        if "putBackItemsFromTrash" in url and method == "POST":
            if resp := self._handle_drive_trash_recover(data):
                return resp

        if "deleteItems" in url and method == "POST":
            if resp := self._handle_drive_trash_delete(data):
                return resp

    def _handle_drive_endpoints_get(
        self, url, method, params
    ) -> Optional[ResponseMock]:
        """Handle drive endpoints get requests."""
        if "com.apple.CloudDocs/download/by_id" in url and method == "GET" and params:
            if resp := self._handle_drive_download(params):
                return resp

        if "icloud-content.com" in url and method == "GET":
            if resp := self._handle_icloud_content(url):
                return resp

        if "/appleauth/auth/authorize/signin" in url and method == "GET":
            return ResponseMock(
                "",
                status_code=200,
                headers={
                    "scnt": "scnt_value",
                    "X-Apple-Session-Id": "session_id_value",
                    "X-Apple-Auth-Attributes": "auth_attributes_value",
                },
            )

    def _handle_setup_endpoint(
        self, url, method, data, headers
    ) -> Optional[ResponseMock]:
        """Handle setup endpoint requests."""
        if "accountLogin" in url and method == "POST":
            return self._handle_account_login(data)

        if "listDevices" in url and method == "GET":
            return ResponseMock(TRUSTED_DEVICES)

        if "sendVerificationCode" in url and method == "POST":
            return self._handle_send_verification_code(data)

        if "validateVerificationCode" in url and method == "POST":
            return self._handle_validate_verification_code(data)

        if "validate" in url and method == "POST" and headers:
            return self._handle_validate(headers)

    def _handle_auth_endpoint(self, url, method, data) -> Optional[ResponseMock]:
        """Handle auth endpoint requests."""
        if "signin" in url and method == "POST":
            return self._handle_signin(data)

        if "securitycode" in url and method == "POST":
            return self._handle_security_code(data)

        if "trust" in url and method == "GET":
            return ResponseMock("", status_code=204)

    def _handle_account_login(self, data: dict[str, Any]) -> ResponseMock:
        """Handle account login."""
        if data.get("dsWebAuthToken") not in VALID_TOKENS:
            self._raise_error(MagicMock(), None, "Unknown reason")
        if data.get("dsWebAuthToken") == REQUIRES_2FA_TOKEN:
            return ResponseMock(LOGIN_2FA)
        return ResponseMock(LOGIN_WORKING)

    def _handle_send_verification_code(self, data: dict[str, Any]) -> ResponseMock:
        """Handle send verification code."""
        if data == TRUSTED_DEVICE_1:
            return ResponseMock(VERIFICATION_CODE_OK)
        return ResponseMock(VERIFICATION_CODE_KO)

    def _handle_validate_verification_code(self, data: dict[str, Any]) -> ResponseMock:
        """Handle validate verification code."""
        TRUSTED_DEVICE_1.update(
            {
                "verificationCode": "0",
                "trustBrowser": True,
            }
        )
        if data == TRUSTED_DEVICE_1:
            self._service._apple_id = AUTHENTICATED_USER
            return ResponseMock(VERIFICATION_CODE_OK)
        self._raise_error(MagicMock(), None, "FOUND_CODE")

    def _handle_validate(self, headers: dict[str, Any]) -> ResponseMock:
        """Handle validate."""
        if headers.get("X-APPLE-WEBAUTH-TOKEN") == VALID_COOKIE:
            return ResponseMock(LOGIN_WORKING)
        self._raise_error(MagicMock(), None, "Session expired")

    def _handle_signin(self, data: dict[str, Any]) -> ResponseMock:
        """Handle signin."""
        if data.get("accountName") not in VALID_USERS:
            self._raise_error(MagicMock(), None, "Unknown reason")
        if data.get("accountName") == REQUIRES_2FA_USER:
            self._service.session._data["session_token"] = REQUIRES_2FA_TOKEN
            return ResponseMock(AUTH_OK)

        self._service.session._data["session_token"] = VALID_TOKEN
        return ResponseMock(AUTH_OK)

    def _handle_security_code(self, data: dict[str, Any]) -> ResponseMock:
        """Handle security code."""
        if data.get("securityCode", {}).get("code") != VALID_2FA_CODE:
            self._raise_error(MagicMock(), None, "Incorrect code")

        self._service.session._data["session_token"] = VALID_TOKEN
        return ResponseMock("", status_code=204)

    def _handle_drive_retrieve(self, data: dict[Any, Any]) -> Optional[ResponseMock]:
        """Handle drive retrieve item details."""
        drivewsid = data[0].get("drivewsid")
        if drivewsid == "FOLDER::com.apple.CloudDocs::root":
            return ResponseMock(DRIVE_ROOT_WORKING)
        if drivewsid == "FOLDER::com.apple.Preview::documents":
            return ResponseMock(DRIVE_ROOT_INVALID)
        if drivewsid == "FOLDER::com.apple.CloudDocs::TRASH_ROOT":
            return ResponseMock(DRIVE_TRASH_WORKING)
        if (
            drivewsid
            == "FOLDER::com.apple.CloudDocs::1C7F1760-D940-480F-8C4F-005824A4E05B"
        ):
            return ResponseMock(DRIVE_FOLDER_WORKING)
        if (
            drivewsid
            == "FOLDER::com.apple.CloudDocs::D5AA0425-E84F-4501-AF5D-60F1D92648CF"
        ):
            return ResponseMock(DRIVE_SUBFOLDER_WORKING)

    def _handle_drive_trash_recover(
        self, data: dict[str, Any]
    ) -> Optional[ResponseMock]:
        """Handle drive trash recover."""
        items_data = data.get("items")
        if (
            items_data
            and items_data[0].get("drivewsid")
            == "FOLDER::com.apple.CloudDocs::2BF8600B-5DCC-4421-805A-1C28D07197D5"
        ):
            return ResponseMock(DRIVE_TRASH_RECOVER_WORKING)

    def _handle_drive_trash_delete(
        self, data: dict[str, Any]
    ) -> Optional[ResponseMock]:
        """Handle drive trash delete forever."""
        items_data = data.get("items")
        if (
            items_data
            and items_data[0].get("drivewsid")
            == "FOLDER::com.apple.CloudDocs::478AEA23-42A2-468A-ABC1-1A04BC07F738"
        ):
            return ResponseMock(DRIVE_TRASH_DELETE_FOREVER_WORKING)

    def _handle_drive_download(self, params: dict[str, Any]) -> Optional[ResponseMock]:
        """Handle drive download."""
        if params.get("document_id") == "516C896C-6AA5-4A30-B30E-5502C2333DAE":
            return ResponseMock(DRIVE_FILE_DOWNLOAD_WORKING)

    def _handle_icloud_content(self, url: str) -> Optional[ResponseMock]:
        """Handle iCloud content."""
        if "Scanned+document+1.pdf" in url:
            return ResponseMock({}, raw=BytesIO(b"PDF_CONTENT"))



"""File service."""

from datetime import datetime
from typing import Any, Optional

from requests import Response

from pyicloud.exceptions import PyiCloudAPIResponseException, PyiCloudServiceUnavailable
from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession


class UbiquityService(BaseService):
    """The 'Ubiquity' iCloud service."""

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        super().__init__(service_root, session, params)

        self._root: Optional["UbiquityNode"] = None

        try:
            self.root
        except PyiCloudAPIResponseException as error:
            if error.code == 503:
                raise PyiCloudServiceUnavailable(error.reason) from error
            raise

    @property
    def root(self) -> "UbiquityNode":
        """Gets the root node."""
        if not self._root:
            self._root = self.get_node(0)
        return self._root

    def get_node_url(self, node_id, variant="item") -> str:
        """Returns a node URL."""
        return f"{self.service_root}/ws/{self.params['dsid']}/{variant}/{node_id}"

    def get_node(self, node_id) -> "UbiquityNode":
        """Returns a node."""
        response: Response = self.session.get(self.get_node_url(node_id))
        return UbiquityNode(self, response.json())

    def get_children(self, node_id) -> list["UbiquityNode"]:
        """Returns a node children."""
        response: Response = self.session.get(self.get_node_url(node_id, "parent"))
        items: list[dict[str, str]] = response.json()["item_list"]
        return [UbiquityNode(self, item) for item in items]

    def get_file(self, node_id, **kwargs) -> Response:
        """Returns a node file."""
        return self.session.get(self.get_node_url(node_id, "file"), **kwargs)

    def __getattr__(self, attr):
        return getattr(self.root, attr)

    def __getitem__(self, key) -> "UbiquityNode":
        return self.root[key]


class UbiquityNode:
    """Ubiquity node."""

    def __init__(self, conn: UbiquityService, data: dict[str, str]) -> None:
        self.data: dict[str, str] = data
        self.connection: UbiquityService = conn

        self._children: Optional[list[UbiquityNode]] = None

    @property
    def item_id(self) -> Optional[str]:
        """Gets the node id."""
        return self.data.get("item_id")

    @property
    def name(self) -> str:
        """Gets the node name."""
        return self.data.get("name", "<unknown>")

    @property
    def type(self) -> str:
        """Gets the node type."""
        return self.data.get("type", "<unknown>")

    @property
    def size(self) -> Optional[int]:
        """Gets the node size."""
        try:
            return int(self.data.get("size", "-1"))
        except ValueError:
            return None

    @property
    def modified(self) -> datetime:
        """Gets the node modified date."""
        return datetime.strptime(self.data.get("modified", ""), "%Y-%m-%dT%H:%M:%SZ")

    def open(self, **kwargs) -> Response:
        """Returns the node file."""
        return self.connection.get_file(self.item_id, **kwargs)

    def get_children(self) -> list["UbiquityNode"]:
        """Returns the node children."""
        if not self._children:
            self._children = self.connection.get_children(self.item_id)
        return self._children

    def dir(self) -> list[str]:
        """Returns children node directories by their names."""
        return [child.name for child in self.get_children()]

    def get(self, name: str) -> "UbiquityNode":
        """Returns a child node by its name."""
        return [child for child in self.get_children() if child.name == name][0]

    def __getitem__(self, key: str) -> "UbiquityNode":
        try:
            return self.get(key)
        except IndexError as i:
            raise KeyError(f"No child named {key} exists") from i

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<{self.type.capitalize()}: '{self}'>"
